# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Example Cython-based computational kernels for evaluating the f() in  w' = f(w, t).

This module provides generic kernels for autonomous linear problems of 1st and 2nd order,
with and without mass matrices on the LHS.

(Autonomous: f = f(w), i.e. f does not explicitly depend on t.)

All matrices in this module are assumed to have C memory layout.

See the code and comments in pydgq/solver/builtin_kernels.pyx for details.
"""

from __future__ import division, print_function, absolute_import

import numpy as np

import pylu.dgesv as dgesv
cimport pylu.dgesv as dgesv_c

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.types import DTYPE


# TODO: add an update hook for updating the matrices (if not constant in time)?
#
# The problem is that in a nonlinear problem, f() is called in the innermost loop:
#
#    timestep (base value of t changes)
#      nonlinear iteration (w changes)
#        sub-evaluation at some t inside the timestep (w changes; actual t used for f() evaluation changes)
#
# so when should the solver call the matrix update methods? (assuming "at every call" is too expensive)


cdef class Linear1stOrderKernel(CythonKernel):
#    cdef DTYPE_t* M
#    cdef DTYPE_t[:,::1] M_arr

    def __init__(self, int n, DTYPE_t[:,::1] M):
        """def __init__(self, int n, DTYPE_t[:,::1] M):

A generic kernel for a linear 1st-order problem.
This is basically just a matrix-vector product.

The problem reads

    w' = M w

Trivial example to invoke this kernel:

    from pydgq.solver.types import DTYPE
    from pydgq.solver.builtin_kernels import Linear1stOrderKernel
    from pydgq.solver.odesolve import ivp

    n   = 3  # number of DOFs in your 1st-order system
    w0  = np.ones( (n,), dtype=DTYPE, order="C" )  # your IC here
    M   = np.eye( n, dtype=DTYPE, order="C" )  # your "M" matrix here
    rhs = Linear1stOrderKernel(n, M)
    ivp( ..., rhs=rhs, w0=w0 )
"""
        # super
        CythonKernel.__init__(self, n)

        self.M_arr = M  # keepalive (FIXME: do we need to .copy() to be sure?)
        self.M = &(self.M_arr[0,0])  # get raw pointer

    cdef void callback(self, RTYPE_t t) nogil:  # t unused in this example
        self.compute(self.w, self.out)

    # Derived classes need this, outputting to a temporary output array,
    # so we have this general version with parametrized input/output.
    #
    # In most problem-specific kernels, we wouldn't need this, and could instead
    # perform the computation in callback(), saving one function call per RHS evaluation.
    #
    cdef void compute(self, DTYPE_t* w_in, DTYPE_t* wp_out) nogil:
        cdef int j, k
        for j in range(self.n):  # row
            # w' = M w
            wp_out[j] = 0.
            for k in range(self.n):  # column
                wp_out[j] += self.M[j*self.n + k] * w_in[k]


cdef class Linear1stOrderKernelWithMassMatrix(Linear1stOrderKernel):
#    cdef DTYPE_t* LU
#    cdef int* p
#    cdef DTYPE_t* wrk
#    cdef DTYPE_t[:,::1] LU_arr
#    cdef int[::1] p_arr
#    cdef DTYPE_t[::1] wrk_arr

    def __init__(self, int n, DTYPE_t[:,::1] M, DTYPE_t[:,::1] A):
        """def __init__(self, int n, DTYPE_t[:,::1] M, DTYPE_t[:,::1] A):

First-order problem that has a nontrivial (but constant-in-time) mass matrix on the LHS:

    A w' = M w

Note that if "A" is small enough to invert, one can instead write

    w' = inv(A) M w

and just use a Linear1stOrderKernel with  inv(A) M  as the matrix.

This class obtains w' by first computing A w', and then solving a linear equation system for w':

    g := M w
    A w' = g

Trivial example to invoke this kernel:

    from pydgq.solver.types import DTYPE
    from pydgq.solver.builtin_kernels import Linear1stOrderKernelWithMassMatrix
    from pydgq.solver.odesolve import ivp

    n   = 3  # number of DOFs in your 1st-order system
    w0  = np.ones( (n,), dtype=DTYPE, order="C" )  # your IC here
    A   = np.eye( n, dtype=DTYPE, order="C" )  # your "A" matrix here
    M   = np.eye( n, dtype=DTYPE, order="C" )  # your "M" matrix here
    rhs = Linear1stOrderKernelWithMassMatrix(n, M, A)
    ivp( ..., rhs=rhs, w0=w0 )
"""
        # super
        Linear1stOrderKernel.__init__(self, n, M)

        # LU decompose mass matrix
        #
        # TODO: need some variant of zgesv if DTYPE is ZTYPE
        #
        self.LU_arr, self.p_arr = dgesv.lup_packed(A)
        self.mincols_arr,self.maxcols_arr = dgesv.find_bands(self.LU_arr, tol=1e-15)
        self.wrk_arr = np.empty( (n,), dtype=DTYPE, order="C" )

        # get raw pointers for C access
        self.LU      = &(self.LU_arr[0,0])
        self.p       = &(self.p_arr[0])
        self.mincols = &(self.mincols_arr[0])
        self.maxcols = &(self.maxcols_arr[0])
        self.wrk     = &(self.wrk_arr[0])

    cdef void callback(self, RTYPE_t t) nogil:  # t unused in this example
        # compute g = M w, save result in self.wrk
        self.compute(self.w, self.wrk)

        # solve linear equation system A w' = g  (g stored in self.wrk, result stored in self.out)
        #
        # TODO: need some variant of zgesv if DTYPE is ZTYPE
        #
        dgesv_c.solve_decomposed_banded_c( self.LU, self.p, self.mincols, self.maxcols, self.wrk, self.out, self.n )


cdef class Linear2ndOrderKernel(CythonKernel):
#    cdef int m
#    cdef DTYPE_t* M0
#    cdef DTYPE_t* M1
#    cdef DTYPE_t[:,::1] M0_arr
#    cdef DTYPE_t[:,::1] M1_arr

    def __init__(self, int n, DTYPE_t[:,::1] M0, DTYPE_t[:,::1] M1):
        """def __init__(self, int n, DTYPE_t[:,::1] M0, DTYPE_t[:,::1] M1):

A generic kernel for the linear 2nd-order problem:

    u'' = M0 u + M1 u'

Following the companion method, we define

    v := u'

obtaining a twice larger 1st-order problem

    v' = M0 u + M1 v
    u' = v

We now define

    w := (u1, v1, u2, v2, ..., um, vm)

where m is the number of DOFs of the original 2nd-order system.

Given w, M0 and M1, this class computes w'.

The parameter n specifies the size of the *1st-order* system; n is always even.

Trivial example to invoke this kernel:

    from pydgq.solver.types import DTYPE
    from pydgq.solver.builtin_kernels import Linear2ndOrderKernel
    from pydgq.solver.odesolve import ivp

    m   = 3        # number of DOFs in your 2nd-order system here
    n   = 2*m      # corresponding number of DOFs in the reduced 1st-order system (always 2*m)
    w0  = np.empty( (n,), dtype=DTYPE, order="C" )
    w0[0::2] = 0.  # your IC on u  here
    w0[1::2] = 1.  # your IC on u' here
    M0  = np.eye( m, dtype=DTYPE, order="C" )  # your "M0" matrix here
    M1  = np.eye( m, dtype=DTYPE, order="C" )  # your "M1" matrix here
    rhs = Linear1stOrderKernel(n, M0, M1)  # NOTE: size parameter is n, not m
    ivp( ..., rhs=rhs, w0=w0 )
"""
        if n % 2 != 0:
            raise ValueError("For a 2nd-order problem reduced to a 1st-order one, n must be even; got %d" % (n))

        # super
        CythonKernel.__init__(self, n)

        self.m = n // 2

        # keepalive (FIXME: .copy() to be sure?)
        self.M0_arr = M0
        self.M1_arr = M1

        # get raw pointers for C access
        self.M0 = &(self.M0_arr[0,0])
        self.M1 = &(self.M1_arr[0,0])

    cdef void callback(self, RTYPE_t t) nogil:  # t unused in this example
        self.compute(self.w, self.out)

    # Derived classes need this, outputting to a temporary output array,
    # so we have this general version with parametrized input/output.
    #
    # In most problem-specific kernels, we wouldn't need this, and could instead
    # perform the computation in callback(), saving one function call per RHS evaluation.
    #
    cdef void compute(self, DTYPE_t* w_in, DTYPE_t* wp_out) nogil:
        cdef int j, k
        for j in range(self.m):  # row
            # u' = v
            wp_out[2*j] = w_in[2*j + 1]

            # v' = M0 u + M1 v
            wp_out[2*j + 1] = 0.
            for k in range(self.m):  # column
                wp_out[2*j + 1] += self.M0[j*self.m + k] * w_in[2*j]  +  self.M1[j*self.m + k] * w_in[2*j + 1]


# cdef classes are single inheritance only, so we have some duplication here
# (since this is both a "linear 2nd-order kernel" as well as a "kernel with mass matrix").
#
cdef class Linear2ndOrderKernelWithMassMatrix(Linear2ndOrderKernel):
#    cdef DTYPE_t* LU
#    cdef int* p
#    cdef DTYPE_t* wrk1
#    cdef DTYPE_t* wrk2
#    cdef DTYPE_t* wrk3
#    cdef DTYPE_t[:,::1] LU_arr
#    cdef int[::1] p_arr
#    cdef DTYPE_t[::1] wrk_arr

    def __init__(self, int n, DTYPE_t[:,::1] M0, DTYPE_t[:,::1] M1, DTYPE_t[:,::1] M2):
        """def __init__(self, int n, DTYPE_t[:,::1] M0, DTYPE_t[:,::1] M1, DTYPE_t[:,::1] M2):

A generic kernel for the linear 2nd-order problem with a nontrivial (but constant-in-time) mass matrix:

    M2 u'' = M0 u + M1 u'

This problem is commonly encountered in mechanics.

Companion form:

    M2 v' = M0 u + M1 u'
       u' = v

The parameter n specifies the size of the *1st-order* system; n is always even.

Note that the kernel needs to solve only an  m x m  linear system for the DOFs representing v';
the other m DOFs (u') are obtained directly.

Trivial example to invoke this kernel:

    from pydgq.solver.types import DTYPE
    from pydgq.solver.builtin_kernels import Linear2ndOrderKernelWithMassMatrix
    from pydgq.solver.odesolve import ivp

    m   = 3        # number of DOFs in your 2nd-order system here
    n   = 2*m      # corresponding number of DOFs in the reduced 1st-order system (always 2*m)
    w0  = np.empty( (n,), dtype=DTYPE, order="C" )
    w0[0::2] = 0.  # your IC on u  here
    w0[1::2] = 1.  # your IC on u' here
    M0  = np.eye( m, dtype=DTYPE, order="C" )  # your "M0" matrix here
    M1  = np.eye( m, dtype=DTYPE, order="C" )  # your "M1" matrix here
    M2  = np.eye( m, dtype=DTYPE, order="C" )  # your "M2" matrix here
    rhs = Linear2ndOrderKernelWithMassMatrix(n, M0, M1, M2)  # NOTE: size parameter is n, not m
    ivp( ..., rhs=rhs, w0=w0 )
"""
        # super
        Linear2ndOrderKernel.__init__(self, n, M0, M1)

        # LU decompose mass matrix
        #
        # TODO: need some variant of zgesv if DTYPE is ZTYPE
        #
        self.LU_arr, self.p_arr = dgesv.lup_packed(M2)
        self.mincols_arr,self.maxcols_arr = dgesv.find_bands(self.LU_arr, tol=1e-15)
        self.wrk_arr = np.empty( (2*n,), dtype=DTYPE, order="C" )  # to avoid unnecessary memory fragmentation, allocate both work arrays as one block

        # get raw pointers for C access
        self.LU      = &(self.LU_arr[0,0])
        self.p       = &(self.p_arr[0])
        self.mincols = &(self.mincols_arr[0])
        self.maxcols = &(self.maxcols_arr[0])
        self.wrk1    = &(self.wrk_arr[0])         # first n elements of work space
        self.wrk2    = &(self.wrk_arr[n])         # next m elements of work space
        self.wrk3    = &(self.wrk_arr[n+self.m])  # last m elements of work space

    cdef void callback(self, RTYPE_t t) nogil:  # t unused in this example
        # compute RHS (i.e. w'), store result in wrk1
        #
        # Note that the RHS is exactly the same as in Linear2ndOrderKernel,
        # only the interpretation of the result differs.
        #
        # Thus we can use super's compute() to evaluate the RHS.
        #
        self.compute(self.w, self.wrk1)

        # reorder DOFs, store result in wrk2
        #
        # We must undo the interleaving  w' = ( u'1, M2 v'1, u'2, M2 v'2,..., u'm, M2 v'm )
        # to use pylu.dgesv on the  M2 v'  DOFs only.
        #
        cdef int j
        for j in range(self.m):
            self.wrk2[j] = self.wrk1[2*j+1]  # DOFs corresponding to M2 v'

        # solve  M2 v' = wrk2  for v', store result in wrk3
        #
        # TODO: need some variant of zgesv if DTYPE is ZTYPE
        #
        dgesv_c.solve_decomposed_banded_c( self.LU, self.p, self.mincols, self.maxcols, self.wrk2, self.wrk3, self.m )

        # write output
        for j in range(self.m):
            self.out[2*j]   = self.wrk1[2*j]  # u'
            self.out[2*j+1] = self.wrk3[j]    # v'




# -*- coding: utf-8 -*-

from __future__ import absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.kernel_interface cimport KernelBase
from pydgq.solver.integrator_interface cimport ImplicitIntegrator

# Base class for Galerkin integrators.
#
# Handles data array allocation and provides helper methods common to all Galerkin integrators.
#
#    # array shapes and types:
#
#    # instance arrays (see galerkin.DataManager.allocate_storage())
#    cdef DTYPE_t[:,:,::1] g     = np.empty( [n_space_dofs,n_time_dofs,n_quad], dtype=DTYPE, order="C" )  # effective load vector, for each space DOF, for each time DOF, at each integration point
#    cdef DTYPE_t[:,::1]   b     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # right-hand sides (integral, over the timestep, of g*psi)
#    cdef DTYPE_t[:,::1]   u     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients (unknowns)
#    cdef DTYPE_t[:,::1]   uprev = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients from previous iteration
#    cdef DTYPE_t[:,::1]   uass  = np.empty( [n_quad,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for integration (this ordering needed for speed!)
#    cdef DTYPE_t[::1]     ucorr = np.empty( [n_quad],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for integration)
#    cdef DTYPE_t[:,::1]   uvis  = np.empty( [nt_vis,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for visualization
#    cdef DTYPE_t[::1]     ucvis = np.empty( [nt_vis],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for visualization)
#
#    # global arrays, same for each solver instance (see galerkin.DataManager.load_data(), galerkin.DataManager.prep_solver())
#    cdef RTYPE_t[:,::1] LU      = galerkin.datamanager.LU       # LU decomposed mass matrix (packed format), for one space DOF, shape (n_time_dofs, n_time_dofs)
#    cdef int[::1]       p       = galerkin.datamanager.p        # row permutation information, length n_time_dofs
#    cdef int[::1]       mincols = galerkin.datamanager.mincols  # band information for L, length n_time_dofs
#    cdef int[::1]       maxcols = galerkin.datamanager.maxcols  # band information for U, length n_time_dofs
#    cdef RTYPE_t[::1]   qw      = galerkin.datamanager.integ_w  # quadrature weights (Gauss-Legendre)
#    cdef RTYPE_t[:,::1] psi     = galerkin.datamanager.integ_y  # basis function values at the quadrature points, psi[j,i] is N[j]( x[i] )
#    cdef RTYPE_t[:,::1] psivis  = galerkin.datamanager.vis_y    # basis function values at the visualization points, psivis[j,i] is N[j]( x[i] )
#
cdef class GalerkinIntegrator(ImplicitIntegrator):
    cdef int n_time_dofs    # how many time DOFs to use for *each* space DOF (this is determined by the choice of q in dG(q))

    # instance arrays (specific to this problem instance)
    cdef DTYPE_t* g
    cdef DTYPE_t* b
    cdef DTYPE_t* u
    cdef DTYPE_t* uprev
    cdef DTYPE_t* uass
    cdef DTYPE_t* ucorr
    cdef DTYPE_t* uvis
    cdef DTYPE_t* ucvis
    # also wrk, but that is already declared in IntegratorBase.

    # global arrays (shared across all problems when using the same settings)

    # dG(q) mass matrix for one space DOF, LU decomposed
    cdef RTYPE_t* LU
    cdef int* p
    cdef int* mincols
    cdef int* maxcols

    cdef int n_quad         # number of integration points in the Gauss-Legendre rule
    cdef RTYPE_t* tquad     # integration points, **scaled to [0,1]**
    cdef RTYPE_t* qw        # quadrature weights
    cdef RTYPE_t* psi       # basis function values at integration points

    cdef int n_vis
    cdef RTYPE_t* tvis      # visualization points (accounting for the interp parameter), **scaled to [0,1]**
    cdef RTYPE_t* psivis    # basis function values at visualization points

    # helper methods:

    # assemble Galerkin series of u at given points
    cdef void assemble( self, RTYPE_t* psi, DTYPE_t* uass, DTYPE_t* ucorr, int n_points ) nogil

    # get the value of u at the end of this timestep
    cdef void final_value( self, DTYPE_t* uass ) nogil

    # integrate a function (provided as values at the quadrature points) over the timestep
    cdef DTYPE_t do_quadrature( self, DTYPE_t* funcvals, RTYPE_t dt ) nogil

# discontinuous Galerkin
#
cdef class DG(GalerkinIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil

# continuous Galerkin
#
cdef class CG(GalerkinIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil




# -*- coding: utf-8 -*-
"""Generate data file (pydgq_data.bin) for pydgq.

The generation is very slow; MPI parallelization is supported to boost performance.

Run this module as the main program (with or without mpiexec) to perform the precalculation.
Command-line options are available; pass the standard --help flag to see them.
"""

from __future__ import division, print_function, absolute_import

import time

try:
    import cPickle as pickle  # Python 2.7
except ImportError:
    import pickle  # Python 3.x

import functools  # reduce (Python 3 compatibility)

import numpy as np

try:
    import mpmath  # Python 3.x
except ImportError:
    import sympy.mpmath as mpmath  # Python 2.7

from pydgq.solver.types import RTYPE  # the precalc data is always real-valued regardless of DTYPE
import pydgq.utils.mpi_shim as mpi_shim
import pydgq.utils.listutils as listutils

# version of precalc.py (this is NOT the version of the pydgq package)
__version__ = "1.0.2"


############################################################################################################
# Worker classes
############################################################################################################

# Memoization
class Cache:
    """Cache for values that have been computed so far. Useful if precomputing for a large number of different divisions of [-1,1] some of which contain identical points.

    The cache is meant to be shared across all instances of Precalc in the current process.

    """
    def __init__(self, q):
        """q = maximum degree (see Precalc)"""

        self.q     = q
        self._data = {}

        # create empty caches for all polynomial degrees
        for j in range(q+1):
            self._data[j] = {}

    # syntactic sugar

    # e.g. "if cache contains 3" --> a sub-cache exists for 3rd degree polynomials
    def __contains__(self, key):
        return key in self._data

    # e.g. "cache[3]" --> the sub-cache for 3rd degree polynomials
    def __getitem__(self, key):
        return self._data[key]

    # the caller is supposed to write only to the individual sub-caches, not directly into the top level one
    def __setitem__(self, key, value):
        raise NotImplementedError("The top-level cache in Cache does not support __setitem__(), maybe you meant to write into cache[j]?")


class Precalc:
    """Precalculate hierarchical (Lobatto) basis functions up to the given degree q (>= 1) at the given points (rank-1 np.array of length >= 1) on the reference element [-1,1]."""

    # Legendre polynomials, high precision from sympy.mpmath to avoid cancellation in hierarchical basis functions
    _P = mpmath.legendre  # lambda j, x: ...

    def __init__(self, q, xx, cache=None):
        """q = maximum degree, xx = vector of points, cache = Cache instance or None

        The cache is used to memoize results for each unique value in xx. This is mainly useful
        for sharing evaluations to future instances, which may get some of the same elements in their xx.
        """

        self.x  = None
        self.y  = None  # "y = f(x)"

        assert(q >= 1)
        assert(np.size(xx) >= 1)

        self.q     = q
        self.xx    = xx
        self.cache = cache

        self._build_hierarchical_basis()

    def _build_hierarchical_basis(self):
        """Build a list of Python functions that evaluate the hierarchical basis functions.

        After running this, self.N[j] is a one-argument lambda that evaluates N_j at x. (Not vectorized!)

        """
        q = self.q

        N = []
        N.append( lambda x: (1./2.) * (1. - x) )  # linear, left endpoint
        N.append( lambda x: (1./2.) * (1. + x) )  # linear, right endpoint

        # bubble functions (see user manual)
        for j in range(2,q+1):
            # HACK: Python 3 compatibility: we must float(j), because some part of the toolchain here wants to convert all arguments to mpf, which does not work for int.
            N.append(  (  lambda j: lambda x : ( self._P(j, x) - self._P(j-2, x) ) / np.sqrt( 2. * (2.*j - 1.) )  )(float(j))  )  # use factory to bind j at define time

        self.N = N

    def run(self):
        """Do the precomputation. MPI-fied."""

        # Create indices and corresponding x values.
        #
        nx    = np.size(self.xx)
        all_i = range(nx)
        all_x = self.xx

        # Divide the work among the tasks.
        #
        split_i = listutils.load_balance_list( all_i, mpi_shim.get_size() )
        my_i    = split_i[ mpi_shim.get_rank() ]

        # Do the local work for task-local items.
        #
        N = self.N
        ly = np.empty( [self.q+1, len(my_i)], dtype=RTYPE )  # local y (basis function values)
        # use caching (if available) to avoid slow re-evaluation of the polynomials for x values already seen.
        if self.cache is not None:
            for li,gi in enumerate(my_i):  # local i, corresponding global i
                x = all_x[gi]
                for j in range(self.q+1):
                    # yes, we use floats as keys, and we do really want floating-point equality down to the last ulp in order to accept the entry as matched.
                    if x in self.cache[j]:
                        ly[j,li] = self.cache[j][x]
                    else:
                        ly[j,li] = N[j]( x )  # Compute. Writing into the ly array forces conversion to RTYPE, as we want.
                        self.cache[j][x] = ly[j,li]

        else:
            for li,gi in enumerate(my_i):  # local i, corresponding global i
                x = all_x[gi]
                for j in range(self.q+1):
                    ly[j,li] = N[j]( x )

        # Gather results.
        #
        if mpi_shim.get_size() > 1:
            # Allocate an array to receive the global data.
            #
            # This will need to be reordered after receiving, because any leftover items (modulo nprocs) are processed by some of the first tasks.
            # We have the correct x indices in split_i; at the first step we just glue together the data arrays in sequence using MPI's Gatherv().
            #
            gshape  = [nx, self.q+1]  # transposed global shape (see below)
            recv    = np.empty( np.prod(gshape), dtype=RTYPE )  # linear buffer for global data (we'll reshape this after receiving)

            counts  = [len(lis)*(self.q+1) for lis in split_i]  # lis contains the task-local item indices; all tasks handle q+1 basis functions
            disps   = [0] + np.cumsum( np.array(counts, dtype=int)[:-1] ).tolist()   #  e.g. [3,3,3,2] -> [0,3,6,9]

            assert( np.prod(np.size(ly)) == counts[mpi_shim.get_rank()] )  # in each task, computed local data size must match actual size of local data

            # The number of basis functions is constant (q+1), while the length of my_i may vary across the tasks.
            #
            # Thus, we transpose the local data array before linearizing it for sending, so that the "slices" are of constant length.
            # This makes it simpler to de-linearize the global array after the gather.
            #
            # For Gatherv(), Allgatherv(), sendbuf, recvbuf, see:
            #    https://wiki.gwdg.de/index.php/Mpi4py
            #
            sendbuf = [ np.reshape(np.transpose(ly),-1), counts[mpi_shim.get_rank()] ]  # local_data, local_count (must be counts[mpi_rank])

            # TODO: other dtypes?
            if RTYPE == np.float64:
                MPI_datatype = mpi_shim.get_mpi().DOUBLE
            elif RTYPE == np.float32:
                MPI_datatype = mpi_shim.get_mpi().SINGLE
            else:
                raise NotImplementedError("Unknown RTYPE %s, cannot transmit data buffer" % (RTYPE))
            recvbuf = [ recv, counts, disps, MPI_datatype ]  # data, counts, displacements, mpi_datatype
            comm = mpi_shim.get_comm_world()
            comm.Allgatherv( sendbuf, recvbuf )

            # De-linearize the received array.
            #
            # This utilizes the fact that in the transposed data, each "item" (one x value for all basis functions) has the same length.
            #
            recv = np.reshape(recv, gshape)

            # Get the permutation of rows that was applied by splitting all_i -> split_i and then gluing together the split_i in sequence.
            #
            # E.g. with nx=11, nprocs=4, we could have
            #
            # MPI rank 0: [0,1,8]
            # MPI rank 1: [2,3,9]
            # MPI rank 2: [4,5,10]
            # MPI rank 3: [6,7]
            #
            # After gather this becomes:
            #
            # [0,1,8,2,3,9,4,5,10,6,7]
            #
            # These are the indices in all_x corresponding to each row in the transposed data array.
            #
            perm = np.array( functools.reduce( lambda x,y: x+y,  split_i ), dtype=int )

            # Get the inverse permutation.
            #
            invperm       = np.empty_like( perm )
            invperm[perm] = np.arange( np.size(perm), dtype=int )  # inverse permutation of range(N): invperm[perm] = range(N)

            # Remap by inverse permutation. This orders the data correctly, so that the ith row of the transposed data corresponds to basis function values at all_x[i].
            #
            # Undo the transpose (applying it again) to obtain the final result. Then self.y[j,i] is N[j]( all_x[i] ).
            #
            self.y = np.transpose( recv[invperm] )
            self.x = all_x

        else:  # only one task, which processes everything
            self.y = ly
            self.x = all_x


############################################################################################################
# Main program
############################################################################################################

def main(q, nx, **kwargs):
    """Create precomputed arrays of basis function values.

    Parameters are the maximum degree q (>= 1) and the maximum number of visualization points nx (>= 1).

    Since the basis is hierarchical, any lower degree is obtained by simply chopping off the extraneous rows, as y[:(desired_q + 1),:].

    Thus, for visualization, different arrays are needed only for different numbers of points, which are taken to be equally spaced
    on the reference element [-1,1]. Arrays are generated for 1, 2, ..., nx points. Caching (memoization) is used to accelerate
    the computation so that only unique points are actually evaluated (e.g. arrays with 3 and 5 points share the point at 0.0)

    For integration, separate arrays are created, of function values at Gauss-Legendre points for integration rules of order 1,2,...,q+1.
    The q+1 rule is sufficient for evaluation of mass matrices of the form N(x)*N(x) (where N has at most degree q).
    The q rule is sufficient for N'(x)*N(x) (needed in dG(q)) and lower orders. (This is assuming an affine coordinate mapping
    from the reference element to the actual timestep, so that the Jacobian of the coordinate mapping is constant on each element.)
    """

    if mpi_shim.get_size() < 2:
        msg = " This script can benefit from MPI parallelization; consider running with 'mpiexec -n <some reasonable number> python -m precalc\n"
        sep = ( "=" * len(msg) ) + "\n"
        mpi_shim.mpi_print("\n\n%s%s%s\n" % (sep, msg, sep))

    data = {}

    # common metadata
    data["maxq"] = q  # note: one less than number of rows in y items

    # memoize for faster computation
    cache = Cache(q)

    #########################################################
    # Create array for visualization (equally spaced points)
    #########################################################

    mpi_shim.mpi_print( "Maximum degree %d (%d basis functions)" % (q, q+1) )

    mpi_shim.mpi_print( "\nGenerating visualization data" )
    mpi_shim.mpi_print( "Equally spaced points: max points = %d" % nx )

    data["vis"] = {}  # data at visualization points
    data["vis"]["maxnx"] = nx

    for k in range(1,nx+1):
        if k == 1:
            # Considering the use of this data in ODE system integration (odesolve.pyx),
            # the end of the timestep is the most reasonable choice in the case of a single point.
            #
            xx = np.array( [1.], dtype=RTYPE )
        else:
            xx = np.linspace( -1., 1., k, dtype=RTYPE )

        mpi_shim.mpi_print( "    Computing for %d points" % k )

        # The cache avoids the need to re-compute the basis functions for any values of x
        # already seen at an earlier iteration of this loop, making this run much faster.
        #
        precalc = Precalc(q, xx, cache)
        precalc.run()

        data["vis"][k] = { "x" : precalc.x, "y" : precalc.y }

#    # NOTE: the data is accessed by indexing by ["vis"][how many points], like this:
#    print data["vis"][3]["x"]  # points (x values)
#    print data["vis"][3]["y"]  # corresponding function values (y values)


    #########################################################
    # Create array for integration (Gauss-Legendre points)
    #########################################################

    mpi_shim.mpi_print( "\nGenerating integrator data (Gauss-Legendre rule)" )

    data["integ"] = {}  # data at Gauss-Legendre integration points
    data["integ"]["maxrule"] = q+1

    for d in range(1,q+2):
        mpi_shim.mpi_print( "    Computing for GL rule of order %d" % d )

        xx,ww = np.polynomial.legendre.leggauss( d )

        # We run this from order 1 up to order q+1, even though the lower-degree rules may not be high enough for exact integration
        # of the highest-degree basis functions, in order to allow for under-integration by the user later.
        #
        # Note that if a lower-degree basis is being used (by chopping off rows of the data array, see above),
        # one can then chop this data, too, using a lower-degree (desired_q + 1) rule to still obtain exact integration.
        #
        # E.g.
        #   o = data["integ"][3]  # Gauss-Legendre data for rule order 3, with basis functions up to degree data["maxq"].
        #                         # This under-integrates most of the higher-degree functions.
        #   b = o["y"][:3,:]  # basis functions 0, 1, 2 only - using the order 3 rule, for these the integration of the mass matrix is exact
        #   x = o["x"]  # the integration points (for information only, not actually needed for computing the integral)
        #   w = o["w"]  # the integration weights
        #
        # Recall that b[j,i] = N_j at x_i.
        #
        # For computing the matrices in Galerkin methods, it is much better to use the analytical results (see legtest3.py) than to use Gauss-Legendre numerically,
        # but the b array is very useful for e.g. evaluating Galerkin series in this basis. Let c be a rank-1 np.array with Galerkin coefficients (in this example, of length 3).
        # Then:
        #
        #   u = np.dot( b, c )
        #
        # gives the values of u at the integration points, where u is a function expressed as a linear combination of the basis functions.
        #
        precalc = Precalc(q, xx, cache)  # the cache likely doesn't help here, but it doesn't hurt to use it, either.
        precalc.run()

        # We save a copy of the Gauss-Legendre weights as "w".
        data["integ"][d] = { "x" : precalc.x, "y" : precalc.y, "w" : ww }

#    # NOTE: the data is accessed by the order of the Gauss-Legendre rule, e.g.:
#    print data["integ"][2]

    #########################################################
    # Save results
    #########################################################

    # In the root process: save results to disk
    #
    if mpi_shim.get_rank() == 0:
        # http://stackoverflow.com/questions/10075661/how-to-save-dictionaries-and-arrays-in-the-same-archive-with-numpy-savez
        with open('pydgq_data.bin', 'wb') as outfile:
            pickle.dump( data, outfile, protocol=pickle.HIGHEST_PROTOCOL )
        print( "Wrote pydgq_data.bin" )

# TODO: how to find the data file in an actual installation? (need to save it relative to the package directory)
# TODO: change to .mat format to make the data file more self-documenting?


############################################################################################################
# Command line parser
############################################################################################################

if __name__ == '__main__':

    # MPI support: only the root process parses command-line arguments
    #
    kwargs = None
    comm = mpi_shim.get_comm_world()
    if mpi_shim.get_rank() == 0:
        # We wrap the parsing in try/finally and always broadcast *something*
        # so that the non-root processes can silently exit if the parsing fails.
        #
        # http://stackoverflow.com/questions/25087360/parsing-arguments-using-argparse-and-mpi4py
        #
        try:
            import argparse
            parser = argparse.ArgumentParser(description="""Precalculate hierarchical (Lobatto) basis functions for Galerkin integrators.

For high degrees, the definition of the bubble functions exhibits numerical cancellation, and must thus be computed at increased precision before casting the result to target precision. This is done using arbitrary-precision floating point math, which relies on a pure software implementation and is thus very slow. This script performs the required precomputation and saves the result to disk.

This script supports MPI for parallelization.""", formatter_class=argparse.RawDescriptionHelpFormatter)

            parser.add_argument( '-v', '--version', action='version', version=('%(prog)s ' + __version__) )

            group_behavior = parser.add_argument_group('behavior', 'Precalculator behavior options.')

            group_behavior.add_argument( '-q', '--degree',
                                         dest='q',
                                         default=10,
                                         type=int,
                                         metavar='n',
                                         help='Sets the highest degree (the "q" in "dG(q)") to precompute. Must be >= 1. (Very high degrees, such as 50, are supported by this software, but usually dG(q) gives the best results for q=1 or q=2.) Default %(default)s.' )

            group_behavior.add_argument( '-nx', '--points',
                                         dest='nx',
                                         default=101,
                                         type=int,
                                         metavar='n',
                                         help='For visualization use: sets how many evenly spaced points (on the reference element [-1,1]) each basis function will be evaluated at. Must be >= 1. Default %(default)s.' )

            # http://parezcoydigo.wordpress.com/2012/08/04/from-argparse-to-dictionary-in-python-2-7/
            kwargs = vars( parser.parse_args() )

        finally:
            # Broadcast to all ranks (even if parsing failed, so that the other ranks won't hang)
            #
            if mpi_shim.get_size() > 1:
                kwargs = comm.bcast(kwargs, root=0)

    else:
        # Other ranks: receive broadcast, and exit if empty
        #
        kwargs = comm.bcast(kwargs, root=0)

        if kwargs is None:
            exit(0)

    main(**kwargs)




# -*- coding: utf-8 -*-
#
# Trying out the NumPy API for Legendre polynomials and Gauss--Legendre quadrature,
# with an eye toward the modern hierarchical (Lobatto) basis functions for Galerkin methods
# (B. Szabó, I. Babuška, Finite element analysis, John Wiley & Sons, 1991).
#
# JJ 2016-02-16

from __future__ import division, print_function, absolute_import

import time

import numpy as np

try:
    import mpmath  # Python 3.x
except ImportError:
    import sympy.mpmath as mpmath  # Python 2.7

import matplotlib.pyplot as plt

import pylu.dgesv as dgesv


class RandomPileOfTestStuff:
    # Create a NumPy wrapper for high-precision Legendre polynomials from mpmath.
    #
    # Note that vectorize() is just a convenience wrapper; the implementation is
    # essentially a Python for loop, so it won't increase performance over manually
    # looping over the items.
    #
    # See
    #   http://docs.scipy.org/doc/numpy-1.10.1/reference/generated/numpy.vectorize.html
    #
    #
    # Static method!
    #
    # (Actually, technically speaking this is a function object bound to the class scope;
    #  see the rules for creating static members and the discussion on Python's class system,
    #   http://stackoverflow.com/questions/3506150/static-class-members-python
    #   https://docs.python.org/2/tutorial/classes.html#class-definition-syntax
    # )
    #
    _P = np.vectorize( mpmath.legendre )

    def __init__(self, q=15):
        assert( q >= 1 )

        # max polynomial degree
        self.q = q

        self.P   = None  # Legendre polynomials
        self.N   = None  # hierarchical basis functions (FEM, dG)

        # Galerkin matrices
        self.K   = None  # N' * N'
        self.C   = None  # N' * N
        self.M   = None  # N  * N

        self.get_legendre_polynomials()
        self.build_hierarchical_basis()
        self.build_K()
        self.build_C()
        self.build_M()

    def get_legendre_polynomials(self):
        q = self.q

        # We use a function factory to freeze j to the given value (i.e. bind it at define time instead of at call time).
        #
        # See
        #   http://stackoverflow.com/questions/1107210/python-lambda-problems
        #
        # Note that in Python, it is valid to call a static method by self.f() (as well as the traditional ClassName.f()).
        #
        P = []
        for j in range(q):
            P.append(  ( lambda j: lambda x : self._P(j, x) )(j)  )

        self.P = P

    def build_hierarchical_basis(self):
        q = self.q

        N = []
        N.append( lambda x: (1./2.) * (1. - x) )
        N.append( lambda x: (1./2.) * (1. + x) )
        # explicit solution, using mpmath's high-precision routines (slow, but works fine at least to q=300)
        for j in range(2,q+1):
            # HACK: Python 3 compatibility: we must float(j), because some part of the toolchain here wants to convert all arguments to mpf, which does not work for int.
            N.append(  (  lambda j: lambda x : ( self._P(j, x) - self._P(j-2, x) ) / np.sqrt( 2. * (2.*j - 1.) )  )(float(j))  )

        self.N = N

    # stiffness matrix, integrand N' * N'
    def build_K(self):
        q = self.q
        n = q+1

        K = np.eye( n, dtype=np.float64 )
        K[0,0] =  1./2.
        K[0,1] = -1./2.
        K[1,0] = -1./2.
        K[1,1] =  1./2.

        self.K = K

    # damping or gyroscopic matrix, integrand N' * N
    def build_C(self):
        q = self.q
        n = q+1

        C = np.zeros( (n,n), dtype=np.float64 )
        C[0,0] = -1./2.
        C[0,1] =  1./2.
        C[1,0] = -1./2.
        C[1,1] =  1./2.

        if q >= 2:
            t = 1./np.sqrt(6.)
            C[0,2] = -t
            C[1,2] =  t
            C[2,0] =  t
            C[2,1] = -t

        # General formula for C_ji for j,i >= 2.
        for j in range(2,n):
            i = j + 1  # i-1 = j  <=>  i = j+1
            if i >= 2 and i < n:
                C[j,i] =  2. * np.sqrt( 1. / ( ( 2.*j - 1. ) * ( 2.*j + 1. ) ) )
            i = j - 1  # i+1 = j  <=>  i = j-1
            if i >= 2 and i < n:
                C[j,j-1] = -2. * np.sqrt( 1. / ( ( 2.*j - 1. ) * ( 2.*j - 3. ) ) )

        self.C = C

    # mass matrix, integrand N * N
    def build_M(self):
        q = self.q
        n = q+1

        M = np.zeros( (n,n), dtype=np.float64 )
        M[0,0] = 2./3.
        M[0,1] = 1./3.
        M[1,0] = 1./3.
        M[1,1] = 2./3.
        if q >= 2:
            t = 1./np.sqrt(2.)
            M[0,2] = -t
            M[1,2] = -t
            M[2,0] = -t
            M[2,1] = -t
        if q >= 3:
            t = 1. / (3. * np.sqrt(10) )
            M[0,3] = -t
            M[1,3] =  t
            M[3,0] = -t
            M[3,1] =  t

        # General formula for M_ji for j,i >= 2.
        for j in range(2,n):
            M[j,j]   = 1. / (2.*j - 1.) * ( 1. / (2.*j + 1.)  +  1. / (2.*j - 3.) )
            i = j - 2  # i+2 = j  <=>  i = j-2
            if i >= 2 and i < n:
                M[j,i] = 1. / ( np.sqrt(2.*j - 5.) * (2.*j - 3.) * np.sqrt(2.*j - 1.) )
            i = j + 2  # i-2 = j  <=>  i = j+2
            if i >= 2 and i < n:
                M[j,i] = 1. / ( np.sqrt(2.*j - 1.) * (2.*j + 1.) * np.sqrt(2.*j + 3.) )

        self.M = M


def main():
    stuff = RandomPileOfTestStuff(q=100)

    # From the API docs for numpy.polynomial.legendre.leggauss:
    #    Computes the sample points and weights for Gauss-Legendre quadrature.
    #    These sample points and weights will correctly integrate polynomials of degree 2*deg - 1 or less over the interval [-1, 1] with the weight function f(x) = 1.
    #
    # Hence, in Galerkin methods, to exactly handle a mass matrix where neither of the terms is differentiated, using affine mapping to the reference element [-1,1]
    # (implying piecewise constant Jacobian), we need to have
    #
    #   2*deg - 1 = 2*d
    #
    # i.e.
    #
    #   deg = (2*d + 1) / 2
    #
    # where d is the degree of the highest-degree polynomial present in the Galerkin basis, and deg is the order of the Gauss-Legendre rule.
    # Obviously, since only integer deg are available, we must round up (if rounded down, 2*deg-1 is less than the smallest needed, 2*d).
    # Thus the actual practical result is
    #
    #   deg = ceil( (2*d + 1) / 2 ) = d + ceil( 1/2 ) = d+1
    #
    # (Observe that a rule of this order can do one degree more than the matrix M (integrand N*N) needs. With this, we could exactly integrate x*N*N, if needed.)
    #
    #
    # For the purposes of solving the first-order problem  u' = f(u, t)  by dG, the matrix is not our M, but instead our C (N'*N = degree d-1 plus degree d), so
    #
    #   2*deg - 1 = 2*d - 1
    #
    # i.e.
    #
    #   deg = d
    #
    # Thus, we can solve this problem with a Gauss-Legendre rule of one order lower than in the case where the matrix M is needed.
    #
    #
#    deg = d+1
#    q,w = np.polynomial.legendre.leggauss( deg )
#    print( deg,(2*deg-1),q,w )

    # matrix, name, figure number to plot, bugcheck (max(abs(bugcheck)) should evaluate to 0 for mat[2:,2:])
    data = ( (stuff.K, "K", 2, lambda v: v - np.transpose(v)),
             (stuff.C, "C", 3, lambda v: v + np.transpose(v)),
             (stuff.M, "M", 4, lambda v: v - np.transpose(v)) )

    for mat,name,figno,bugcheck in data:
        print( mat )
        plt.figure(figno)

        plt.subplot(1,2, 1)

        plt.spy(mat)  # spy() doesn't work for a full matrix without any zero entries! (try stuff.M with q=2)

#        plt.imshow(mat, interpolation="nearest", cmap="Oranges")
#        plt.colorbar()

        plt.plot( [0,stuff.q], [0,stuff.q], 'r--' )  # mark diagonal
        plt.title(r"$\mathbf{%s}$" % name)

        if stuff.q >= 2:
            v = mat[2:,2:]
            b = np.max( np.abs( bugcheck(v) ) )
            assert b == 0.0, "bugcheck fail for matrix %s; should be 0, got %g" % (name, b)


        # LU decomposition (sort of)
        #
        plt.subplot(1,2, 2)
        A = mat.copy()
        A[0,0] += 1.0  # K and C are rank-deficient by one; simulate effect of boundary conditions (or dG jump term)
        LU,p = dgesv.lup_packed(A)
        plt.spy(LU)
        plt.plot( [0,stuff.q], [0,stuff.q], 'r--' )
        plt.title(r"$\mathbf{LU}$ (packed format)")


##    L,U,p = dgesv.lup(stuff.M)
##    print( np.transpose(np.nonzero(L)) )
##    print( np.transpose(np.nonzero(U)) )
##    print( p )
##    plt.figure(3)
##    plt.subplot(1,2, 1)
##    plt.spy(L)
##    plt.plot( [0,stuff.q-1], [0,stuff.q-1], 'r--' )
###    plt.imshow(L, interpolation="nearest", cmap="Oranges")
###    plt.colorbar(orientation="horizontal")
##    plt.title(r"$\mathbf{L}$")
##    plt.subplot(1,2, 2)
##    plt.spy(U)
##    plt.plot( [0,stuff.q-1], [0,stuff.q-1], 'r--' )
###    plt.imshow(U, interpolation="nearest", cmap="Oranges")
###    plt.colorbar(orientation="horizontal")
##    plt.title(r"$\mathbf{U}$")


##    LU,p = dgesv.lup_packed(stuff.M)
##    plt.figure(4)
##    plt.spy(LU)
##    plt.plot( [0,stuff.q-1], [0,stuff.q-1], 'r--' )
##    plt.title(r"$\mathbf{LU}$ (packed format)")

##    mincols,maxcols = dgesv.find_bands(LU, 1e-15)
##    print( mincols, maxcols )


##    # old Python-based mincols, maxcols finding code
##
##    # Find the smallest column index with nonzero data on each row in L.
##    #
##    # We can use this to "sparsify" the backsolve even though the data structure is dense.
##    #
##    # This assumes that each row has at least one nonzero entry (which is always the case for an invertible matrix).
##    #
##    Lnz = np.nonzero(L)
##    mincols = []
##    rowprev = -1
##    n = len(Lnz[0])
##    i = 0
##    while i < n:
##        if Lnz[0][i] != rowprev:
##            mincols.append(Lnz[1][i])
##            rowprev = Lnz[0][i]
##        i += 1
##    mincols = np.array( mincols, dtype=np.intc, order="C" )
##    print( L )
##    print( mincols )

##    # Find the largest column index with nonzero data on each row in U.
##    #
##    # We can use this to "sparsify" the backsolve even though the data structure is dense.
##    #
##    # This assumes that each row has at least one nonzero entry (which is always the case for an invertible matrix).
##    #
##    Unz = np.nonzero(U)
##    maxcols = []
##    rowprev = -1
##    n = len(Unz[0])
##    i = n - 1
##    while i >= 0:
##        if Unz[0][i] != rowprev:
##            maxcols.append(Unz[1][i])
##            rowprev = Unz[0][i]
##        i -= 1
##    maxcols.reverse()
##    maxcols = np.array( maxcols, dtype=np.intc, order="C" )
##    print( U )
##    print( maxcols )


    # Visualize
    #
    xx = np.linspace(-1., 1., 101)
    plt.figure(1)
    plt.clf()
    for func in stuff.N:
        plt.plot( xx, func(xx) )

    plt.axis('tight')
    a = plt.axis()
    plt.axis( [ a[0], a[1], a[2]*1.05, a[3]*1.05 ] )

    plt.grid(b=True, which='both')
    plt.title('Hierarchical basis functions')


# naive solve (repeat the LU decomposition process each time)
#
def method1(reps, A, b, x):
    for j in range(reps):
#        dgesv.solve( A, b[j,:], x )
        dgesv.solve( A, b, x )

# decompose once, then solve
#
def method2(reps, A, b, x):
    LU,p = dgesv.lup_packed(A)
    for j in range(reps):
#        dgesv.solve_decomposed( LU, p, b[j,:], x )
        dgesv.solve_decomposed( LU, p, b, x )

# decompose once, then solve, utilize banded structure
#
def method3(reps, A, b, x):
    LU,p = dgesv.lup_packed(A)
    mincols,maxcols = dgesv.find_bands(LU, tol=1e-15)
    for j in range(reps):
#        dgesv.solve_decomposed_banded( LU, p, mincols, maxcols, b[j,:], x )
        dgesv.solve_decomposed_banded( LU, p, mincols, maxcols, b, x )


class MyTimer:
    t0 = None
    l  = None
    n  = None  # number of runs

    def __init__(self, label="", n=None):
        self.label = label
        self.n     = n

    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, type, value, traceback):
        dt  = time.time() - self.t0
        l   = ("%s: " % self.label) if len(self.label) else "time taken: "
        avg = (", avg. %gs per run" % (dt/self.n)) if self.n is not None else ""
        print( "%s%gs%s" % (l, dt, avg) )


if __name__ == '__main__':
    main()
    plt.show()

#    # Running the benchmark loop at the Python end makes the banded version look slower (for our matrix "C", the C code is actually ~3x faster than the generic non-banded version),
#    # because a large majority of the execution time is taken up by data conversion from Python to C and back (and Python asserts, if enabled).
#    #
#    # For small matrices (q = 15 or so), to get reliable results on the C code only (which is a realistic use case if used from inside a Cython-accelerated solver, which is the whole point of dgesv.pyx),
#    # the benchmark looping must be done inside dgesv.pyx.
#    #
#    # Naive benchmarking results start becoming reliable around q >= 100.
#    #
#    reps   = 10000
##    qstart = 3
##    qend   = 16  # actually one-past-end
#    qstart = 300
#    qend   = 301

#    for q in range(qstart, qend):
#        stuff = RandomPileOfTestStuff(q)
#        A = stuff.M

#        n = np.shape(A)[0]
##        A[0,0] += 1.0  # simulate effect of dG jump term (this is the only basis function which is nonzero at x = -1)
##        b = np.random.uniform(0.0, 1.0, size=(reps,n))  # this makes slicing part of the performance measurement - not good
#        b = np.random.uniform(0.0, 1.0, size=(n,))
#        x = np.empty( [n], dtype=np.float64, order="C" )

#        print( "Timings for %d runs" % reps )
##        with MyTimer("%dx%d naive" % (n,n), reps) as mt:
##            print( mt.label )
##            method1(reps, A, b, x)
#        with MyTimer("%dx%d decompose-once" % (n,n), reps) as mt:
#            method2(reps, A, b, x)
#        with MyTimer("%dx%d decompose-once-banded" % (n,n), reps) as mt:
#            method3(reps, A, b, x)
#        print( "Residual from last run %g" % np.max(np.abs( np.dot(A,x) - b )) )




# -*- coding: utf-8 -*-
#
# Usage example: solve the Lorenz system using a custom Python kernel.

from __future__ import division, print_function

import numpy as np

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D

from pydgq.solver.types import DTYPE
from pydgq.solver.galerkin import init
from pydgq.solver.kernel_interface import PythonKernel
import pydgq.solver.odesolve
from pydgq.utils.discontify import discontify  # for plotting dG results


#####################
# config
#####################

q = 2  # degree of basis for dG and cG

# How many visualization (interpolation) points to use within each timestep for Galerkin methods.
#
# Note that the dG solution has the best accuracy at the endpoint of the timestep;
# to compare apples-to-apples with classical integrators, this should be set to 1.
#
# Larger values (e.g. 11) are useful for visualizing the behavior of the dG solution inside
# the timestep (something the classical integrators do not model at all).
#
nt_vis_galerkin = 11

nt = 3500  # number of timesteps
dt = 0.1   # timestep size

save_from = 0  # see pydgq.solver.odesolve.ivp()


#####################
# custom kernel
#####################

# A kernel for the Lorenz system.
#
# The custom kernel only needs to override callback(); even __init__ is not strictly needed,
# unless adding some custom parameters (like here).
#
class LorenzKernel(PythonKernel):
    def __init__(self, rho, sigma, beta):
        # super
        PythonKernel.__init__(self, n=3)

        # custom init
        self.rho   = rho
        self.sigma = sigma
        self.beta  = beta

    def callback(self, t):
        # dxdt = sigma (y - x)
        # dydt = x (rho - z) - y
        # dzdt = x y - beta z
        self.out[0] = self.sigma * (self.w[1] - self.w[0])
        self.out[1] = self.w[0]  * (self.rho  - self.w[2]) - self.w[1]
        self.out[2] = self.w[0]  * self.w[1] - self.beta * self.w[2]  # this is nonlinear, so we can't use a built-in linear kernel


#####################
# main program
#####################

def test(integrator, nt_vis):
    n_saved_timesteps = pydgq.solver.odesolve.n_saved_timesteps( nt, save_from )
    result_len        = pydgq.solver.odesolve.result_len( nt, save_from, interp=nt_vis )
    startj,endj       = pydgq.solver.odesolve.timestep_boundaries( nt, save_from, interp=nt_vis )

    n  = 3  # the Lorenz system has 3 DOFs

    # we use the same values as in the example at https://en.wikipedia.org/wiki/Lorenz_system
    #
    rho   = 28.
    sigma = 10.
    beta  = 8./3.

    # set IC
    #
    w0 = np.empty( (n,), dtype=DTYPE, order="C" )
    w0[0] = 0.
    w0[1] = 2.
    w0[2] = 20.

    # instantiate kernel
    rhs = LorenzKernel(rho=rho, sigma=sigma, beta=beta)

    # create output arrays
    ww   = None #np.empty( (result_len,n), dtype=DTYPE, order="C" )    # result array for w; if None, will be created by ivp()
    ff   = np.empty( (result_len,n), dtype=DTYPE, order="C" )          # optional,  result array for w', could be None
    fail = np.empty( (n_saved_timesteps,), dtype=np.intc, order="C" )  # optional,  fail flag for each timestep, could be None

    # solve problem
    ww,tt = pydgq.solver.odesolve.ivp( integrator=integrator, allow_denormals=False,
                                       w0=w0, dt=dt, nt=nt,
                                       save_from=save_from, interp=nt_vis,
                                       rhs=rhs,
                                       ww=ww, ff=ff, fail=fail,
                                       maxit=10 )

    # visualize
    #
    # http://matplotlib.org/mpl_toolkits/mplot3d/tutorial.html
    #
    print( "** Plotting solution **" )
    fig = plt.figure(1)
    plt.clf()

    # Axes3D has a tendency to underestimate how much space it needs; it draws its labels
    # outside the window area in certain orientations.
    #
    # This causes the labels to be clipped, which looks bad. We prevent this by creating the axes
    # in a slightly smaller rect (leaving a margin). This way the labels will show - outside the Axes3D,
    # but still inside the figure window.
    #
    # The final touch is to set the window background to a matching white, so that the
    # background of the figure appears uniform.
    #
    fig.patch.set_color( (1,1,1) )
    fig.patch.set_alpha( 1.0 )
    x0y0wh = [ 0.02, 0.02, 0.96, 0.96 ]  # left, bottom, width, height      (here as fraction of subplot area)

    ax = Axes3D(fig, rect=x0y0wh)

    # show the discontinuities at timestep boundaries if using dG (and actually have something to draw within each timestep)
    if integrator == "dG" and nt_vis > 1:
        tt = discontify( tt, endj - 1, fill="nan" )

        wtmp = np.empty( (tt.shape[0],n), dtype=DTYPE, order="C" )
        for j in range(n):
            # we need the copy() to get memory-contiguous data for discontify() to process
            wtmp[:,j] = discontify( ww[:,j].copy(), endj - 1, fill="nan" )

        ax.plot( wtmp[:,0], wtmp[:,1], wtmp[:,2], linewidth=0.5  )
    else:
        ax.plot( ww[:,0], ww[:,1], ww[:,2], linewidth=0.5  )

    plt.grid(b=True, which="both")
    plt.axis("tight")
    ax.set_xlabel(r"$x$")
    ax.set_ylabel(r"$y$")
    ax.set_zlabel(r"$z$")
    plt.suptitle(r"Lorenz system: $\rho = %g$, $\sigma = %g$, $\beta = %g$, $x_0 = %g$, $y_0 = %g$, $z_0 = %g$" % (rho, sigma, beta, w0[0], w0[1], w0[2]))

if __name__ == '__main__':
    print("** Solving the Lorenz system **")

    nt_vis = nt_vis_galerkin
    init(q=q, method="dG", nt_vis=nt_vis, rule=None)
    test(integrator="dG", nt_vis=nt_vis)
    plt.show()




# -*- coding: utf-8 -*-

from __future__ import division, print_function, absolute_import

from numpy import float64, complex128, nan

# real
RTYPE = float64
RNAN  = nan

# complex
ZTYPE = complex128
ZNAN  = nan * (1. + 1j)  # both real and imaginary parts nan (to avoid pitfalls for things like isnan(imag(w)))

# problem data (choose real or complex here)
DTYPE = RTYPE
DNAN  = RNAN  # appropriate NaN value for DTYPE




# -*- coding: utf-8 -*-

from __future__ import absolute_import

# real (choose precision here)
ctypedef double RTYPE_t

# complex (choose precision here)
ctypedef double complex ZTYPE_t

# problem data (choose real or complex here)
ctypedef RTYPE_t DTYPE_t




# -*- coding: utf-8 -*-
#
# Example Cython-based computational kernels for evaluating the f() in  w' = f(w, t).
#
# This module provides generic kernels for autonomous linear problems of 1st and 2nd order,
# with and without mass matrices on the LHS.
#
# (Autonomous: f = f(w), i.e. f does not explicitly depend on t.)
#
# All matrices in this module are assumed to have C memory layout.

from __future__ import division, print_function, absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.kernel_interface cimport CythonKernel


# w' = M w
#
# The matrix M is supplied to __init__ (def method, not declared here).
# but callback() is free to do whatever it wants with it.
#
# The data attributes self.timestep and self.iteration (in ancestor) are provided
# for user callback()s to keep track of whether they need to update something.
#
# This class assumes M is constant in time.
#
cdef class Linear1stOrderKernel(CythonKernel):
    cdef DTYPE_t* M
    cdef DTYPE_t[:,::1] M_arr

    cdef void callback(self, RTYPE_t t) nogil
    cdef void compute(self, DTYPE_t* w_in, DTYPE_t* wp_out) nogil


# A w' = M w
#
# with the mass matrix "A" constant with respect to time.
#
# Same note about "M" as above.
#
cdef class Linear1stOrderKernelWithMassMatrix(Linear1stOrderKernel):
    cdef DTYPE_t* LU
    cdef int* p
    cdef int* mincols
    cdef int* maxcols
    cdef DTYPE_t* wrk
    cdef DTYPE_t[:,::1] LU_arr
    cdef int[::1] p_arr
    cdef int[::1] mincols_arr
    cdef int[::1] maxcols_arr
    cdef DTYPE_t[::1] wrk_arr

    cdef void callback(self, RTYPE_t t) nogil


# u'' = M0 u + M1 u'
#
# reduced to the twice larger 1st-order system
#
#  u' = v
#  v' = M0 u + M1 v
#
# The callback() method computes w',
# where w = (u1, v1, u2, v2, ..., um, vm).
#
# The above note about "M" applies to "M0" and "M1".
#
# Note that  n = 2*m  still denotes the size of the **1st-order** system.
#
cdef class Linear2ndOrderKernel(CythonKernel):
    cdef int m  # size of original 2nd-order system, m = n/2
    cdef DTYPE_t* M0
    cdef DTYPE_t* M1
    cdef DTYPE_t[:,::1] M0_arr
    cdef DTYPE_t[:,::1] M1_arr

    cdef void callback(self, RTYPE_t t) nogil
    cdef void compute(self, DTYPE_t* w_in, DTYPE_t* wp_out) nogil


# M2 u'' = M0 u + M1 u'
#
# reduced to
#
#     u' = v
#  M2 v' = M0 u + M1 u'
#
# The callback() method computes w',
# where w = (u1, v1, u2, v2, ..., um, vm).
#
# with the mass matrix "M2" constant with respect to time.
#
# Same note about "M0" and "M1" as above.
#
# Note that  n = 2*m  still denotes the size of the **1st-order** system.
#
cdef class Linear2ndOrderKernelWithMassMatrix(Linear2ndOrderKernel):
    # cdef classes are single inheritance only, so we have some duplication here
    # (since this is both a "linear 2nd-order kernel" as well as a "kernel with mass matrix").
    cdef DTYPE_t* LU
    cdef int* p
    cdef int* mincols
    cdef int* maxcols
    cdef DTYPE_t* wrk1  # n elements
    cdef DTYPE_t* wrk2  # m elements
    cdef DTYPE_t* wrk3  # m elements
    cdef DTYPE_t[:,::1] LU_arr
    cdef int[::1] p_arr
    cdef int[::1] mincols_arr
    cdef int[::1] maxcols_arr
    cdef DTYPE_t[::1] wrk_arr  # wrk1, wrk2, wrk3 all stored here

    cdef void callback(self, RTYPE_t t) nogil




# -*- coding: utf-8 -*-

from __future__ import absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.kernel_interface cimport KernelBase

# Base class for all integrators.
#
# Do not inherit directly from this!
#
# Inherit from ExplicitIntegrator or ImplicitIntegrator depending on your algorithm.
#
cdef class IntegratorBase:
    cdef str name              # human-readable name of algorithm (set by derived classes): "RK4", "dG", etc.
    cdef KernelBase rhs        # RHS computational kernel (problem-specific, passed in by user)

    cdef DTYPE_t[::1] wrk_arr  # work space (to be allocated by derived classes)
    cdef DTYPE_t* wrk          # raw C pointer to wrk_arr (to be filled in by derived classes)

    # call interface for solver
    #
    # w  : in/out: old state vector -> new state vector
    # t  : in: time at the beginning of the timestep (passed through to self.rhs.call())
    # dt : in: size of timestep to take
    #
    # return value: number of implicit solve iterations taken for this timestep (explicit integrators must always return 1)
    #
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil

# Base class for explicit integrators.
#
cdef class ExplicitIntegrator(IntegratorBase):
    pass  # no new data attributes or cdef methods

# Base class for implicit integrators.
#
cdef class ImplicitIntegrator(IntegratorBase):
    cdef int maxit  # maximum number of Banach/Picard iterations in implicit solve




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Interface for integration algorithms.

These base classes provide the interface that connects the IVP solver
with the implementations of the specific algorithms (RK4, dG, etc.).

Beyond instantiation of the objects, this interface is only available
at the Cython level.

See the source code in pydgq.solvers.integrator_interface.pyx for details.
"""

from __future__ import absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.kernel_interface cimport KernelBase

cdef class IntegratorBase:
    def __init__(self, str name, KernelBase rhs):
        """def __init__(self, str name, KernelBase rhs):

Base class for all integration algorithms.

Do not inherit directly from this!

Inherit from ExplicitIntegrator or ImplicitIntegrator depending on your algorithm.

Parameters:
    name : str
        Human-readable algorithm name such as "RK4", "dG", etc.

        This is intended to be provided by __init__ of a concrete
        derived class that implements a specific algorithm.

        See pydgq.solver.explicit, pydgq.solver.implicit and
        pydgq.solver.galerkin for examples.

    rhs : KernelBase instance (in practice, an instance of a derived class)
        The RHS computational kernel.

        This is intended to be supplied by the user at problem solve time,
        so derived classes should also keep rhs as a parameter to __init__.

Data attributes:
    name : str
        saved from __init__
    rhs : KernelBase instance
        saved from __init__
    wrk_arr : DTYPE_t[::1]
        work space for integration algorithm, underlying NumPy array
        (set to None in __init__; actual array to be allocated by derived classes)
    wrk : DTYPE_t*
        raw C pointer to data in wrk_arr, only visible at Cython level
        (set to NULL in __init___; to be filled in by derived classes)

No further docstrings are provided, because the rest of this class
is not visible from the Python level. See the source code in
pydgq.solvers.integrator_interface.pyx for details.

Basically, when implementing a new algorithm, the derived cdef class
should override the method:

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:

where the parameters are:
    w  : in/out
        old state vector in -> new state vector out
    t  : in
        time at the beginning of the timestep (passed through to self.rhs.call()
        to support RHS that explicitly depend on t)
    dt : in
        size of timestep to take

and the return value is:
    int :
        number of implicit solve iterations taken for this timestep.
        Explicit integrators must always return 1.
"""
        self.name = name
        self.rhs  = rhs

        # work array not created by default
        self.wrk_arr = None
        self.wrk = <DTYPE_t*>0

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        return 0  # default implementation: no-op, no iterations taken


cdef class ExplicitIntegrator(IntegratorBase):
    def __init__(self, str name, KernelBase rhs):
        """def __init__(self, str name, KernelBase rhs):

Base class for explicit integrators (such as RK4).

Inheriting from this class simply marks the algorithm
as an explicit one.

No new parameters or data attributes are added here.
"""
        # super
        IntegratorBase.__init__(self, name, rhs)


# TODO/FIXME: currently each integrator needs to implement the iteration loop separately.
cdef class ImplicitIntegrator(IntegratorBase):
    def __init__(self, str name, KernelBase rhs, int maxit):
        """def __init__(self, str name, KernelBase rhs, int maxit):

Base class for implicit integrators (such as IMR and dG).

Note that inheriting from this class simply marks the algorithm
as an implicit one; each integrator needs to implement the
nonlinear iteration loop separately.

Parameters added by ImplicitIntegrator:
    maxit : int
        Maximum number of Banach/Picard iterations to take
        in the nonlinear iteration loop.

Data attributes added by ImplicitIntegrator:
    maxit : int
        saved from __init__
"""
        # super
        IntegratorBase.__init__(self, name, rhs)

        self.maxit = maxit




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Integrators based on the Galerkin method, to update w by one timestep.

Also provided is a DataManager class that loads in the data generated by pydgq.utils.precalc,
required to use classes derived from GalerkinIntegrator.

Do not instantiate DataManager directly; instead, call the init() function exported by this module.
"""
# TODO: might be cleaner if the user could instantiate DataManager directly (would trivially allow multiple instances with different settings)

from __future__ import division, print_function, absolute_import

import sys

try:
    import cPickle as pickle  # Python 2.7
except ImportError:
    import pickle  # Python 3.x

import pkg_resources  # for accessing installed data file (pydgq_data.bin)

import numpy as np

import  pylu.dgesv as dgesv    # Cython-based LU decomposition and linear equation system solver, callable from inside nogil blocks -- Python interface (using np.arrays)
cimport pylu.dgesv as dgesv_c  # -"- -- Cython interface (using raw pointers, explicit sizes)

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.types import DTYPE, RTYPE
from pydgq.solver.kernel_interface cimport KernelBase
from pydgq.solver.integrator_interface cimport ImplicitIntegrator
from pydgq.solver.compsum cimport accumulate


cdef class GalerkinIntegrator(ImplicitIntegrator):

    def __init__(self, str name, KernelBase rhs, int maxit):
        """def __init__(self, str name, KernelBase rhs, int maxit):

Base class for integrators based on the Galerkin method.

Handles data array allocation and provides helper methods common to all Galerkin integrators.

Parameters:
    as in ancestor (pydgq.solver.integrator_interface.ImplicitIntegrator).
"""
        global datamanager

        if datamanager is None:
            raise RuntimeError("%s: init() must be called first." % name)
        if not datamanager.available:
            raise RuntimeError("%s: Cannot use Galerkin integrators because the auxiliary class did not initialize." % name)
        if datamanager.method != name:
            raise RuntimeError("%s: Trying to integrate with %s, but the auxiliary class has been initialized for %s." % (name, name, datamanager.method))

        # super
        ImplicitIntegrator.__init__(self, name, rhs, maxit)

        self.n_time_dofs  = datamanager.n_time_dofs
        self.n_quad       = datamanager.rule   # number of quadrature points (Gauss-Legendre integration points)
        self.n_vis        = datamanager.nt_vis

        # allocate instance-specific arrays
        #
        my_storage = datamanager.allocate_storage(solver_id=id(self), n_space_dofs=rhs.n)

        # retrieve instance arrays
        #
        cdef DTYPE_t[:,:,::1] g          = my_storage["g"]      # effective load vector, for each space DOF, for each time DOF, at each integration point
        cdef DTYPE_t[:,::1]   b          = my_storage["b"]      # right-hand sides (integral, over the timestep, of g*psi)
        cdef DTYPE_t[:,::1]   u          = my_storage["u"]      # Galerkin coefficients (unknowns)
        cdef DTYPE_t[:,::1]   uprev      = my_storage["uprev"]  # Galerkin coefficients from previous iteration
        cdef DTYPE_t[:,::1]   uass       = my_storage["uass"]   # u, assembled for integration
        cdef DTYPE_t[::1]     ucorr      = my_storage["ucorr"]  # correction for compensated summation in assemble() (for integration)
        cdef DTYPE_t[:,::1]   uvis       = my_storage["uvis"]   # u, assembled for visualization
        cdef DTYPE_t[::1]     ucvis      = my_storage["ucvis"]  # correction for compensated summation in assemble() (for visualization)
        cdef DTYPE_t[::1]     wrk_arr    = my_storage["wrk"]    # work space for dG(), cG()

        # get raw pointers to array data (needed by our C routines)
        #
        # FIXME/TODO: This is brittle, since the array shape information is carried separately from the data; and only implicitly
        # FIXME/TODO: by convention between this class and DataManager. Could also store the arrays here?
        #
        self.g       = &g[0,0,0]
        self.b       = &b[0,0]
        self.u       = &u[0,0]
        self.uprev   = &uprev[0,0]
        self.uass    = &uass[0,0]
        self.ucorr   = &ucorr[0]
        self.uvis    = &uvis[0,0]
        self.ucvis   = &ucvis[0]
        self.wrk     = &wrk_arr[0]

        # retrieve global arrays
        #
        # NOTE: our "t" = "x" in precalc.py; it is a general support module for Lobatto basis functions, not specific to its ODE solver use case here.
        #
        cdef RTYPE_t[:,::1]   LU         = datamanager.LU       # LU decomposed mass matrix (packed format), for one space DOF, shape (n_time_dofs, n_time_dofs)
        cdef int[::1]         p          = datamanager.p        # row permutation information, length n_time_dofs
        cdef int[::1]         mincols    = datamanager.mincols  # band information for L, length n_time_dofs
        cdef int[::1]         maxcols    = datamanager.maxcols  # band information for U, length n_time_dofs
        cdef RTYPE_t[::1]     qw         = datamanager.integ_w  # quadrature weights (Gauss-Legendre)
        cdef RTYPE_t[:,::1]   psi        = datamanager.integ_y  # basis function values at the quadrature points, qy[j,i] is N[j]( tquad[i] )
        cdef RTYPE_t[:,::1]   psivis     = datamanager.vis_y    # basis function values at the visualization points, qy[j,i] is N[j]( tvis[i] )
        cdef RTYPE_t[::1]     tvis       = datamanager.vis_x    # time values at the visualization points (on the reference element [-1,1])
        cdef RTYPE_t[::1]     tquad      = datamanager.integ_x  # Gauss-Legendre points of the chosen rule, in (-1,1)

        # get raw pointers to array data
        #
        self.LU      = &LU[0,0]
        self.p       = &p[0]
        self.mincols = &mincols[0]
        self.maxcols = &maxcols[0]
        self.qw      = &qw[0]
        self.psi     = &psi[0,0]
        self.psivis  = &psivis[0,0]
        self.tvis    = &tvis[0]
        self.tquad   = &tquad[0]

        # map tvis and tquad: [-1,1] --> [0,1] (reference element --> offset within timestep)
        cdef unsigned int j
        for j in range(self.n_vis):
            tvis[j] += 1.0
            tvis[j] *= 0.5
        for j in range(self.n_quad):
            tquad[j] += 1.0
            tquad[j] *= 0.5

    def __del__(self):
        global datamanager
        datamanager.free_storage(solver_id=id(self))


    # Assemble Galerkin series at given points inside the timestep.
    #
    # The points are specified implicitly, by evaluating the basis functions at the desired points (see the parameter psi, below).
    # This is parametrized to allow for different use cases (integration, and visualization/reporting).
    #
    # Space DOFs are independent, each has its own Galerkin series. This assembles them all in one go.
    #
    # psi:    rank-2 array, size [n_time_dofs, n_points]:     values of basis functions, evaluated at each of the points t[i] (at which the series is to be assembled); psi[j,i] is psi[j]( t[i] )
    # uass:   rank-2 array, size [n_points, n_space_dofs]:    output, values of u; uass[j,i] is u_i( t[j] )
    # ucorr:  rank-1 array, size [n_points]:                  work area for correction terms in compensated summation
    #
    # Additionally, this automatically uses self.u:
    #
    # self.u: rank-2 array, size [n_space_dofs, n_time_dofs]: Galerkin coefficients; u[j,i] is the coefficient of psi[i] for solution component u_j(t)
    #
    #                         basis         output         summ. corr. wrk
    cdef void assemble( self, RTYPE_t* psi, DTYPE_t* uass, DTYPE_t* ucorr, int n_points ) nogil:
        cdef unsigned int i,j,k
        cdef int n_space_dofs = self.rhs.n
        cdef int n_time_dofs  = self.n_time_dofs

#        # naive summation
#        for k in range(n_space_dofs):
#    
#            # zero out the assembled values for this space DOF
#            for i in range(n_points):
#                uass[ i*n_space_dofs + k ] = 0.0
#    
#            # sum contributions from each basis function
#            for j in range(n_time_dofs):
#                for i in range(n_points):
#                    uass[ i*n_space_dofs + k ] += self.u[ k*n_time_dofs + j ] * psi[ j*n_points + i ]

        # compensated summation
        cdef DTYPE_t* ps
        cdef DTYPE_t* pc
        for k in range(n_space_dofs):

            # Zero out the assembled values for this space DOF.
            #
            # Also zero out the correction array. Note that since we process one space DOF at a time, we only need one correction array.
            # The same position uass[i,k] gets written multiple times during the inner loops, but each time k changes, a new summation begins
            # (at that point, the sums for any previous k are not updated any more).
            #
            for i in range(n_points):
                uass[ i*n_space_dofs + k ] = 0.0
                ucorr[ i ] = 0.0

            # Sum contributions from each basis function.
            for i in range(n_points):
                # Compute the pointers only once for each i.
                ps = &uass[ i*n_space_dofs + k ]  # sum (accumulator)
                pc = &ucorr[ i ]                  # correction
                for j in range(n_time_dofs):
                    accumulate( ps, pc, self.u[ k*n_time_dofs + j ] * psi[ j*n_points + i ] )


    # Assemble Galerkin series at the end of the timestep.
    #
    # uass:   rank-1 array, size [n_space_dofs]:    output, values of u; uass[i] is u_i( t )
    #
    # Compared to the general-purpose assemble():
    #   - Only N[1] (the linear basis function associated with the endpoint) is nonzero at the endpoint.
    #   - Its value there is exactly 1, so we can just take the corresponding Galerkin coefficient from u[] as the function value
    #     (no need to actually sum anything).
    #   - There are always at least 2 basis functions (the linear ones), so we don't need to check whether this basis function exists.
    #   - Here n_points is always 1.
    #
    cdef void final_value( self, DTYPE_t* uass ) nogil:
        cdef unsigned int k
        cdef int n_space_dofs = self.rhs.n
        cdef int n_time_dofs  = self.n_time_dofs
        for k in range(n_space_dofs):
            # here only one visualization point, so uass[ n_space_dofs*0 + k ] --> uass[ k ]
            uass[ k ] = self.u[ k*n_time_dofs + 1 ]


    # Integrate a function over the timestep by applying a quadrature.
    #
    # Compensated summation is used to obtain maximal accuracy.
    #
    # funcvals:    array of values of the function to be integrated, evaluated at the quadrature points (rank-1 or C order, length self.n_quad)
    # dt:          size of timestep (will be used for scaling the result)
    #
    # Additionally, this uses:
    #
    # self.qw:     array of quadrature weights (rank-1 or C order); length of funcvals must match length of self.qw
    # self.n_quad: number of items in funcvals and qw
    #
    cdef DTYPE_t do_quadrature( self, DTYPE_t* funcvals, RTYPE_t dt ) nogil:
        cdef unsigned int i

#        # naive summation (not good for problems sensitive to initial conditions, such as vibration problems with low damping)
#        cdef DTYPE_t s
#        s = 0.0
#        for i in range(self.n_quad):
#            s += self.qw[i]*funcvals[i]

        # compensated summation
        cdef DTYPE_t s, c
        s = self.qw[0]*funcvals[0]  # first term in sum; no correction by Kahan algorithm (if we wanted to correct, we should take the roundoff from this multiplication, instead)
        c = 0.0
        for i in range(1,self.n_quad):
            accumulate( &s, &c, self.qw[i]*funcvals[i] )

        # Scale the result to account for the length of the timestep.
        #
        # This amounts to a change of variable in the integral, as is commonly used in finite element methods.
        # ( For a refresher, see https://en.wikipedia.org/wiki/Integration_by_substitution#Substitution_for_single_variable )
        #
        # We use a simple affine coordinate mapping so that the Jacobian is constant across the timestep.
        #
        # The reference interval is [-1,1], so its length is 2; the target length is dt; thus we must scale by dt/2.
        #
        # Alternatively, formally speaking, the affine mapping of tau: [-1,1] --> t: [a,b] is given by
        #
        #    t = a + (b - a) ( tau - (-1) ) / ( 1 - (-1) )
        #      = a + (b - a) ( tau + 1 ) / 2
        #      = a + (1/2) (b - a) ( tau + 1 )
        #
        # Hence by differentiating both sides with respect to tau,
        #
        #    dt/dtau = (b - a) / 2
        #
        # Noting that in our application, (b - a) = dt, we obtain the result.
        #
        s *= dt/2.0

        return s


cdef class DG(GalerkinIntegrator):
    def __init__(self, KernelBase rhs, int maxit):
        """def __init__(self, KernelBase rhs, int maxit):

Time-discontinuous Galerkin.
"""
        # super
        GalerkinIntegrator.__init__(self, name="dG", rhs=rhs, maxit=maxit)

    # Integrate w over one timestep.
    #
    # When call() returns, the Galerkin coefficients in self.u correspond to the
    # behavior of w over the last processed timestep. This is used in visualization
    # to support "interp".
    #
    #    # array shapes and types:
    #
    #    # instance arrays (see galerkin.DataManager.allocate_storage())
    #    cdef DTYPE_t[:,:,::1] g     = np.empty( [n_space_dofs,n_time_dofs,n_quad], dtype=DTYPE, order="C" )  # effective load vector, for each space DOF, for each time DOF, at each integration point
    #    cdef DTYPE_t[:,::1]   b     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # right-hand sides (integral, over the timestep, of g*psi)
    #    cdef DTYPE_t[:,::1]   u     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients (unknowns)
    #    cdef DTYPE_t[:,::1]   uprev = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients from previous iteration
    #    cdef DTYPE_t[:,::1]   uass  = np.empty( [n_quad,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for integration (this ordering needed for speed!)
    #    cdef DTYPE_t[::1]     ucorr = np.empty( [n_quad],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for integration)
    #    cdef DTYPE_t[:,::1]   uvis  = np.empty( [nt_vis,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for visualization
    #    cdef DTYPE_t[::1]     ucvis = np.empty( [nt_vis],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for visualization)
    #
    #    # global arrays, same for each solver instance (see galerkin.DataManager.load_data(), galerkin.DataManager.prep_solver())
    #    cdef RTYPE_t[:,::1] LU      = galerkin.datamanager.LU       # LU decomposed mass matrix (packed format), for one space DOF, shape (n_time_dofs, n_time_dofs)
    #    cdef int[::1]       p       = galerkin.datamanager.p        # row permutation information, length n_time_dofs
    #    cdef int[::1]       mincols = galerkin.datamanager.mincols  # band information for L, length n_time_dofs
    #    cdef int[::1]       maxcols = galerkin.datamanager.maxcols  # band information for U, length n_time_dofs
    #    cdef RTYPE_t[::1]   qw      = galerkin.datamanager.integ_w  # quadrature weights (Gauss-Legendre)
    #    cdef RTYPE_t[:,::1] psi     = galerkin.datamanager.integ_y  # basis function values at the quadrature points, psi[j,i] is N[j]( tquad[i] )
    #    cdef RTYPE_t[:,::1] psivis  = galerkin.datamanager.vis_y    # basis function values at the visualization points, psivis[j,i] is N[j]( tvis[i] )
    #
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef DTYPE_t* up = self.wrk  # temporary storage for u'
        cdef unsigned int nequals  # for convergence check

        cdef int n_space_dofs = self.rhs.n
        cdef int n_time_dofs  = self.n_time_dofs
        cdef int n_quad       = self.n_quad

        cdef RTYPE_t tcurr  # t at current quadrature point

        # Loop counters.
        #
        # j = space DOF
        # k = time DOF
        # l = quadrature point
        # m = implicit iteration (Banach fixed point iteration)
        #
        cdef int j, k, l, m=-1

        # Initialize the Galerkin coefficients u: initially, we set u = u0, which is available in the array w.
        #
        # u :     current value of Galerkin coefficients
        # uprev : value of Galerkin coefficients at previous iteration, for convergence check
        #
        for j in range(n_space_dofs):
            # linears
            self.u[j*n_time_dofs + 0]     = w[j]  # value at start of timestep
            self.u[j*n_time_dofs + 1]     = w[j]  # value at end of timestep
            self.uprev[j*n_time_dofs + 0] = w[j]
            self.uprev[j*n_time_dofs + 1] = w[j]

            # bubbles
            for k in range(2,n_time_dofs):
                self.u[j*n_time_dofs + k]     = 0.0
                self.uprev[j*n_time_dofs + k] = 0.0

        # Implicit iteration (Banach/Picard fixed point iteration)
        #
        for m in range(self.maxit):
            self.rhs.begin_iteration(m)  # inform RHS kernel that a new iteration starts

            # Assemble u at the quadrature points for this iteration.
            #
            # This forms the core of the Banach fixed point iteration: here we use the
            # previous Galerkin coefficients u (latest known iterate) to model the behavior of u
            # inside the timestep.
            #
            # The old coefficients are then no longer needed; they get overwritten by the solver below.
            #
            self.assemble( self.psi, self.uass, self.ucorr, n_quad )

            # Form the right-hand side at each quadrature point.
            #
            for l in range(n_quad):
                tcurr = t + dt*self.tquad[l]

                # Get the instantaneous value of u' at this quadrature point, store result in "up".
                #
                self.rhs.call(&self.uass[n_space_dofs*l + 0], up, tcurr)

                # At the quadrature point l, compute the value of the RHS integrand
                # in the weak form (original RHS multiplied by test function).
                #
                # We do this for each space DOF independently.
                #
                # The time DOFs here correspond to the set of test functions.
                #
                for j in range(n_space_dofs):
                    for k in range(n_time_dofs):
                        # g: [n_space_dofs,n_time_dofs,n_quad], values of the RHS integrand at the quadrature points
                        # psi: [n_time_dofs, n_quad], values of the test functions at the quadrature points
                        self.g[j*(n_time_dofs*n_quad) + k*n_quad + l] = up[j] * self.psi[k*n_quad + l]

            # Solve the dG linear equation system (separately for each space DOF).
            #
            for j in range(n_space_dofs):

                # For this space DOF, integrate the right-hand side across the timestep (for each test function).
                #
                # We use the Gauss-Legendre rule that was set up by init(). (This information is implicit
                # in the array qw, and in the array psi, which was above used to evaluate the integrand.)
                #
                for k in range(n_time_dofs):
                    # b: [n_space_dofs,n_time_dofs], load vector for dG linear equation system
                    # g: [n_space_dofs,n_time_dofs,n_quad], values of the RHS integrand at the quadrature points
                    self.b[j*n_time_dofs + k] = self.do_quadrature( &self.g[j*(n_time_dofs*n_quad) + k*n_quad + 0], dt )

                # Add the RHS contribution from the jump term (see user manual). This accounts for the initial condition for this timestep.
                #
                self.b[j*n_time_dofs + 0] += w[j]

                # Solve the dG linear equation system for this space DOF.
                #
                # This updates the Galerkin coefficients self.u.
                #
                # TODO: need some variant of zgesv if DTYPE is ZTYPE
                #
                dgesv_c.solve_decomposed_banded_c( self.LU, self.p, self.mincols, self.maxcols, &self.b[j*n_time_dofs + 0], &self.u[j*n_time_dofs + 0], n_time_dofs )

            # Check convergence; break early if converged to within machine precision.
            #
            nequals = 0
            for j in range(n_space_dofs):
                for k in range(n_time_dofs):
                    if self.u[j*n_time_dofs + k] == self.uprev[j*n_time_dofs + k]:
                        nequals += 1
            if nequals == n_space_dofs*n_time_dofs:
                break

            # Store the current values for performing the convergence check at the next iteration.
            #
            for j in range(n_space_dofs):
                for k in range(n_time_dofs):
                    self.uprev[j*n_time_dofs + k] = self.u[j*n_time_dofs + k]

        # Store into w the final value at the end of this timestep.
        #
        self.final_value( w )

        return (m+1)


cdef class CG(GalerkinIntegrator):
    def __init__(self, KernelBase rhs, int maxit):
        """def __init__(self, KernelBase rhs, int maxit):

Time-continuous Galerkin.

This is almost the same code as dG, the only difference being in the handling of the load vector.

(Be aware that usually dG gives better results.)
"""
        # super
        GalerkinIntegrator.__init__(self, name="cG", rhs=rhs, maxit=maxit)

    # Integrate w over one timestep.
    #
    # When call() returns, the Galerkin coefficients in self.u correspond to the
    # behavior of w over the last processed timestep. This is used in visualization
    # to support "interp".
    #
    #    # array shapes and types:
    #
    #    # instance arrays (see galerkin.DataManager.allocate_storage())
    #    cdef DTYPE_t[:,:,::1] g     = np.empty( [n_space_dofs,n_time_dofs,n_quad], dtype=DTYPE, order="C" )  # effective load vector, for each space DOF, for each time DOF, at each integration point
    #    cdef DTYPE_t[:,::1]   b     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # right-hand sides (integral, over the timestep, of g*psi)
    #    cdef DTYPE_t[:,::1]   u     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients (unknowns)
    #    cdef DTYPE_t[:,::1]   uprev = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients from previous iteration
    #    cdef DTYPE_t[:,::1]   uass  = np.empty( [n_quad,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for integration (this ordering needed for speed!)
    #    cdef DTYPE_t[::1]     ucorr = np.empty( [n_quad],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for integration)
    #    cdef DTYPE_t[:,::1]   uvis  = np.empty( [nt_vis,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for visualization
    #    cdef DTYPE_t[::1]     ucvis = np.empty( [nt_vis],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for visualization)
    #
    #    # global arrays, same for each solver instance (see galerkin.DataManager.load_data(), galerkin.DataManager.prep_solver())
    #    cdef RTYPE_t[:,::1] LU      = galerkin.datamanager.LU       # LU decomposed mass matrix (packed format), for one space DOF, shape (n_time_dofs, n_time_dofs)
    #    cdef int[::1]       p       = galerkin.datamanager.p        # row permutation information, length n_time_dofs
    #    cdef int[::1]       mincols = galerkin.datamanager.mincols  # band information for L, length n_time_dofs
    #    cdef int[::1]       maxcols = galerkin.datamanager.maxcols  # band information for U, length n_time_dofs
    #    cdef RTYPE_t[::1]   qw      = galerkin.datamanager.integ_w  # quadrature weights (Gauss-Legendre)
    #    cdef RTYPE_t[:,::1] psi     = galerkin.datamanager.integ_y  # basis function values at the quadrature points, psi[j,i] is N[j]( tquad[i] )
    #    cdef RTYPE_t[:,::1] psivis  = galerkin.datamanager.vis_y    # basis function values at the visualization points, psivis[j,i] is N[j]( tvis[i] )
    #
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef DTYPE_t* up = self.wrk  # temporary storage for u'
        cdef unsigned int nequals  # for convergence check

        cdef int n_space_dofs = self.rhs.n
        cdef int n_time_dofs  = self.n_time_dofs
        cdef int n_quad       = self.n_quad

        cdef RTYPE_t tcurr  # t at current quadrature point

        # Loop counters.
        #
        # j = space DOF
        # k = time DOF
        # l = quadrature point
        # m = implicit iteration (Banach fixed point iteration)
        #
        cdef int j, k, l, m=-1

        # Initialize the Galerkin coefficients u: initially, we set u = u0, which has been saved as the array w.
        #
        for j in range(n_space_dofs):
            self.u[j*n_time_dofs + 0]     = w[j]  # value at start of timestep
            self.u[j*n_time_dofs + 1]     = w[j]  # value at end of timestep

            # value of Galerkin coefficients at previous iteration, for convergence check
            self.uprev[j*n_time_dofs + 0] = w[j]
            self.uprev[j*n_time_dofs + 1] = w[j]

            # bubbles
            for k in range(2,n_time_dofs):
                self.u[j*n_time_dofs + k]     = 0.0
                self.uprev[j*n_time_dofs + k] = 0.0

        # Implicit iteration (Banach/Picard fixed point iteration)
        #
        for m in range(self.maxit):
            self.rhs.begin_iteration(m)  # inform RHS kernel that a new iteration starts

            # Assemble latest known u at the quadrature points for this iteration.
            #
            self.assemble( self.psi, self.uass, self.ucorr, n_quad )

            # Form the right-hand side at each quadrature point.
            #
            for l in range(n_quad):
                tcurr = t + dt*self.tquad[l]

                # Get the instantaneous value of u' at this quadrature point, store result in "up".
                #
                self.rhs.call(&self.uass[n_space_dofs*l + 0], up, tcurr)

                # Compute the value of the RHS integrand in the weak form.
                for j in range(n_space_dofs):
                    for k in range(1,n_time_dofs):  # note that in cG, the load vector for time DOF 0 will be filled in separately (initial condition)
                        # g: [n_space_dofs,n_time_dofs,n_quad], values of the RHS integrand at the quadrature points
                        # psi: [n_time_dofs, n_quad], values of the test functions at the quadrature points
                        self.g[j*(n_time_dofs*n_quad) + k*n_quad + l] = up[j] * self.psi[k*n_quad + l]

            # Solve the cG linear equation system (separately for each space DOF).
            #
            for j in range(n_space_dofs):

                # For this space DOF, integrate the right-hand side across the timestep (for each test function).
                #
                for k in range(1,n_time_dofs):  # note that in cG, the load vector for time DOF 0 will be filled in separately (initial condition)
                    # b: [n_space_dofs,n_time_dofs], load vector for dG linear equation system
                    # g: [n_space_dofs,n_time_dofs,n_quad], values of the RHS integrand at the quadrature points
                    self.b[j*n_time_dofs + k] = self.do_quadrature( &self.g[j*(n_time_dofs*n_quad) + k*n_quad + 0], dt )

                # Account for the initial condition in the load vector (see user manual).
                #
                # The first time basis function (index 0) is the only one that is nonzero at the start point of the timestep;
                # its value there is exactly 1. Hence the value of the first time DOF (index 0) for space DOF j can be set to w[j].
                #
                # The mass matrix on the LHS has already been adjusted in init() to have the multiplicative identity on the first row,
                # so that when the linear equation system is solved, u[j,0] will just pick up the value set here.
                #
                self.b[j*n_time_dofs + 0] = w[j]

                # Solve the cG linear equation system for this space DOF.
                #
                # This updates the Galerkin coefficients self.u.
                #
                # TODO: need some variant of zgesv if DTYPE is ZTYPE
                #
                dgesv_c.solve_decomposed_banded_c( self.LU, self.p, self.mincols, self.maxcols, &self.b[j*n_time_dofs + 0], &self.u[j*n_time_dofs + 0], n_time_dofs )

            # Check convergence; break early if converged to within machine precision.
            #
            nequals = 0
            for j in range(n_space_dofs):
                for k in range(n_time_dofs):
                    if self.u[j*n_time_dofs + k] == self.uprev[j*n_time_dofs + k]:
                        nequals += 1
            if nequals == n_space_dofs*n_time_dofs:
                break

            # Store the current values for performing the convergence check at the next iteration.
            #
            for j in range(n_space_dofs):
                for k in range(n_time_dofs):
                    self.uprev[j*n_time_dofs + k] = self.u[j*n_time_dofs + k]

        # Store into w the final value at the end of this timestep.
        #
        self.final_value( w )

        return (m+1)


#######################
# Python-level helpers
#######################

# When instantiated, the following object performs the common initialization that is needed only once per process,
# including the construction of the mass matrix for the first-order initial value problem
# for a single space DOF:
#
#   u'    = f( u(t), t ),   t in (-1,1)
#   u(-1) = u0
#
# This is done in generic form over the reference element [-1,1]. The LU decomposition
# of the matrix is also formed. The LU decomposition can then be used for fast solving
# of the linear equation systems in the main solver loop.
#
#
# Thread-safety:
#   Multiple dG (or cG) solvers may be running at the same time, but they *must* have the same settings (method and q).
#   Simultaneous processing with different settings is currently not supported and will cause errors and/or crashes. (TODO)
#
#
# Multiple "space" degrees of freedom in the semilinear ODE system
#
#   M u' = g( u(t), t )                         (*)
#
# are to be handled by the user-provided kernel by formally inverting M:
#
#   u' = M^(-1) g( u(t), t ) =: f( u(t), t )    (**)
#
# The problem (**) (first and last form) is of the standard form handled by the solver.
#
# From the last two forms of (**), we have (multiplying from the left by M)
#
#   M f = g
#
# Thus, the kernel can solve a linear equation system (different from the dG(q) one!) to
# obtain the effective load f whenever it is needed. See kernels.pyx for examples.
#
# This needs M, g and the latest u (all "space" degrees of freedom), and outputs the
# corresponding value of f (for all "space" degrees of freedom).
#
# This in turn allows us to solve the time evolution of each "space" degree of freedom separately,
# for each of them using the time integrator that handles a single "space" degree of freedom.
#
# (In addition to "space" degrees of freedom, i.e. degrees of freedom in the ODE system,
#  there are also "time" degrees of freedom, of which there are q+1 for *each* "space" degree of freedom.
#  These "time" degrees of freedom are the ones whose behavior is encoded in the Galerkin mass matrix C, below.
#
#  We use the label "space" degrees of freedom, because ODE systems like (*) typically arise
#  in space discretization of PDE problems.)
#
#
class DataManager:
    """Helper class to manage the Galerkin solver."""

    def __init__(self, q, method, nt_vis, rule=None):
        """def __init__(self, q, method, nt_vis, rule=None):

Perform global initialization of the Galerkin solver.

This object must be instantiated before using the integrators "dG" or "cG".
See the module-level global function init(), which does this.

During one session (run of the process), one initialization is enough for all simulations
using the same method (combination of "q" and "method"). To change the settings, simply initialize again.

The Galerkin solver uses the hierarchical basis functions (linears, plus bubbles
based on integrated Legendre polynomials).

Parameters:
    q : int, >= 1
        Polynomial degree of basis. In practice, 1 or 2 are usually good values.

        The upper limit available depends on data stored in the precalculated data file (pydgq_data.bin).
        To determine what your data file has, instantiate a DataManager and call its load_data() method.

    method : string
        Which Galerkin method to use:

         "dG" : time-discontinuous Galerkin (recommended)
            The solution is finitely discontinuous (C^{-1}) across timestep boundaries,
            with left-continuity at each boundary.

            Because no degrees of freedom are wasted on enforcing continuity,
            this allows better approximation at the same q.

         "cG" : continuous Galerkin
            Conforming finite elements in time.

            The solution is C0-continuous across timestep boundaries. This may be
            more intuitive for visualization, but note that usually the approximation
            is much worse than with dG.

    nt_vis : int
        Number of visualization points per computed timestep.

        This corresponds to the parameter "interp" of odesolve.result_len(),
        odesolve.make_tt() and odesolve.ivp().

        !!! Currently, the value of interp in all subsequent calls to odesolve.ivp() must be the same
            as the value of nt_vis passed here. This is a implementation-technical limitation
            that may be removed in the future. !!!

    rule : int or None
        Order of the Gauss-Legendre quadrature to use for integrating the load vector.
        If None (default), the highest available order is used (see pydqg.utils.precalc).

        To determine what your data file has, instantiate a DataManager and call its load_data() method.
"""

        if q < 1:
            raise ValueError( "q must be >= 1; got %d" % (q) )
        if nt_vis < 1:
            raise ValueError( "nt_vis must be >= 1; got %d" % (nt_vis) )
        if method not in ["dG", "cG"]:
            raise ValueError( "Unknown method '%s'; valid: 'dG', 'cG'" % (method) )

        self.method      = method
        self.q           = q       # degree of basis
        self.n_time_dofs = q+1     # the hierarchical basis of degree q has q+1 degrees of freedom.
        self.nt_vis      = nt_vis  # number of visualization points per timestep

        try:
            self.__load_data(rule)
            self.__build_C()
            self.__prep_solver()
            self.__storage = {}  # problem instance specific data will be stored here (key = solver id)
            self.available = True
        except Exception as e:
            # NOTE: our "t" = "x" in precalc.py; it is a general support module for Lobatto basis functions, not specific to its ODE solver use case here.

            # precomputed data for Gauss-Legendre integration (used for processing the load vector in the solver)
            self.integ_x     = None   # GL points
            self.integ_y     = None   # values of basis functions at GL points
            self.integ_w     = None   # GL weights
            self.rule        = None   # order of the chosen GL rule (integer)

            # precomputed data for timestep visualization
            self.vis_x       = None   # visualization points (time values on reference element [-1,1])
            self.vis_y       = None   # values of basis functions at visualization points
            self.nt_vis      = None   # number of visualization points within each timestep

            self.q           = None   # degree of basis
            self.n_time_dofs = None   # number of "time degrees of freedom" for each "space degree of freedom"

            self.C           = None   # Galerkin mass matrix
            self.LU          = None   # the LU decomposition of C
            self.p           = None   # row permutation for load vector ( L U = P C,  so if  C u = b,  then  P C u = P b  and thus  L U u = P b )
            self.mincols     = None   # band information for L
            self.maxcols     = None   # band information for U

            self.__storage     = None   # storage for solver instance arrays

            self.available   = False  # Galerkin integrators enabled/disabled

            print( "    ERROR: Initialization failed. Galerkin integrators not available.", file=sys.stderr )
            raise

    # Load precalculated data for the hierarchical basis functions.
    #
    def __load_data(self, rule):
        """Load precalculated data (pydgq_data.bin).

The hierarchical basis functions are numerically sensitive to cancellation especially at high degrees q,
hence must be computed using higher internal precision (double is not enough).

Because arbitrary precision floating point in pure software is slow, pydqg.utils.precalc makes
a precomputed array for this routine to load.

Parameters:
    rule : int or None
        Order of the Gauss-Legendre quadrature to use for integrating the load vector.
        None = highest available.
"""
        q           = self.q
        n_time_dofs = self.n_time_dofs
        nt_vis      = self.nt_vis

        # first try ./pydgq.data.bin (local override)
        data = None
        filename = 'pydgq_data.bin'
        try:
            with open(filename, 'rb') as infile:
                data = pickle.load( infile )
        except IOError as e:
            pass

        # next try ~/.config/pydgq/pydgq_data.bin (user override)
        if data is None:
            import os.path
            home = os.path.expanduser("~")
            url = os.path.join(home, ".config", "pydgq", filename)

            try:
                with open(url, 'rb') as infile:
                    data = pickle.load( infile )
            except IOError as e:
                pass

        # finally, use pkg_resources to find installed data file
        if data is None:
            if not pkg_resources.resource_exists("pydgq", "pydgq_data.bin"):
                raise IOError("Cannot find data file '%s'. The file can be generated by running the module pydqg.utils.precalc as the main program." % (filename))

            try:
                data = pickle.load( pkg_resources.resource_stream("pydgq", "pydgq_data.bin") )
            except:
                print("Cannot find data file '%s'. The file can be generated by running the module pydqg.utils.precalc as the main program." % (filename), file=sys.stderr)
                raise

        maxq      = data["maxq"]
        maxnt_vis = data["vis"]["maxnx"]
        maxrule   = data["integ"]["maxrule"]

        # save maximum values available in data file, for information
        #
        self.maxq      = maxq
        self.maxnt_vis = maxnt_vis
        self.maxrule   = maxrule

        # Sanity check that we have enough data available
        #
        if q > maxq:
            raise ValueError("Data file '%s' contains data up to degree maxq=%d, but the requested degree was higher, q=%d. Use a lower q or re-run pydqg.utils.precalc for higher maxq." % (filename, maxq, q))
        need_nt_vis = nt_vis+1 if self.method == "cG" else nt_vis  # "visualization" points are used for evaluating the result inside the timestep (see parameter "interp" of odesolve.result_len() and odesolve.ivp()).
        if need_nt_vis > maxnt_vis:
            raise ValueError("Data file '%s' contains data up to visualization maxnt_vis=%d, but the needed number of points is higher, %d. Use a lower nt_vis or re-run pydqg.utils.precalc for higher maxnx (same thing as our maxnt_vis)." % (filename, maxnt_vis, need_nt_vis))
        if rule is not None and rule > maxrule:
            raise ValueError("Data file '%s' contains data for integration rules up to order %d, but the requested order was higher, rule=%d. Use a lower rule." % (filename, maxrule, rule))

        # Choose the order of the Gauss-Legendre quadrature and the number of visualization points.
        #
        # The load is not a polynomial, so we default to the highest available integration rule.
        #
        # The Gauss-Legendre rule of order q (where q us the degree of the basis) would be enough for our mass matrix C;
        # the fact that the matrix has the integrand N'*N and not N*N saves us one order, compared to the usual q+1.
        #
        # But we handle the matrix analytically, so the required order needed by the matrix does not matter;
        # Gauss-Legendre is used only for processing the load vector.
        #
        # If we wanted to be exact, we should least-squares fit the load to the given basis before performing the quadrature; this would give us computable error.
        # However, doing that is expensive, since the quadrature runs inside the loop for implicit iterations (iterative linearization using the Banach fixed point theorem).
        #
        self.rule = rule if rule is not None else maxrule
        rule = self.rule

        # Choose the appropriate data arrays
        #
        integ = data["integ"][rule]

        if self.method == "dG":
            # In dG, the value of the solution at the end of the timestep is different
            # from its value at the beginning of the next timestep, even though these points
            # occur at the same value of t.
            #
            # (Technically speaking, the solution is left-continuous, i.e. the end-of-timestep
            #  point has the actual value of the solution at that value of t.
            #
            #  The start-of-next-timestep point represents the limit from the right.)
            #
            # Thus, we just take the subdivision that has nt_vis points (including both endpoints).
            #
            vis = data["vis"][nt_vis]

        else: # self.method == "cG":
            # In cG, the endpoint of timestep n is the start point of timestep n+1.
            # Avoid visualizing it twice.
            #
            # To get the desired number of points, take one more, and then discard the start point,
            # obtaining nt_vis unique points.
            #
            vis = data["vis"][nt_vis+1]
            vis["x"] = vis["x"][1:]
            vis["y"] = vis["y"][:,1:]

        # Chop off higher degree (>q) basis functions from the y arrays, if the arrays contain more data than what we actually need.
        #
        if data["maxq"] > q:
            integ["y"] = integ["y"][ :(q+1), : ]
            vis["y"]   = vis["y"][   :(q+1), : ]

        # NOTE: our "t" = "x" in precalc.py; it is a general support module for Lobatto basis functions, not specific to its ODE solver use case here.
        #
        cdef RTYPE_t[::1]   integ_x = integ["x"].copy(order="C")  # ...and hope for the correct dtype (our RTYPE) in the data file
        cdef RTYPE_t[::1]   integ_w = integ["w"].copy(order="C")
        cdef RTYPE_t[:,::1] integ_y = integ["y"].copy(order="C")
        cdef RTYPE_t[::1]   vis_x   = vis["x"].copy(order="C")
        cdef RTYPE_t[:,::1] vis_y   = vis["y"].copy(order="C")

        self.integ_x = integ_x
        self.integ_y = integ_y
        self.integ_w = integ_w
        self.vis_x   = vis_x
        self.vis_y   = vis_y

        # We extracted what we need, unload the rest.
        #
        del data

        # Now the loaded data can be accessed as follows:
        #
        # Gauss-Legendre integration (will be used for processing the load vector):
        #  - self.integ_x: Gauss-Legendre points of the chosen rule, in (-1,1)
        #  - self.integ_w: corresponding Gauss-Legendre weights
        #  - self.integ_y: values of basis functions at the Gauss-Legendre points; integ_y[j,i] is N[j]( tquad[i] )
        #  - self.rule:    order of the chosen Gauss-Legendre rule (for information only)
        #
        # Visualization:
        #  - self.vis_x:   x values for visualization (time value on reference element [-1,1])
        #  - self.vis_y:   values of basis functions at the visualization points; vis_y[j,i] is N[j]( tvis[i] )

    def __build_C(self):
        """Build the dG(q) mass matrix C.

C contains integrals of Ni'*Nj (i column, j row) over the reference element [-1,1],
where the Ns are the hierarchical (Lobatto) basis functions.

This routine uses an analytical result (see user manual), avoiding the need
to numerically evaluate the integrands to get this. This approach is both
much more accurate and much faster.
"""
        # Note that there is no need to scale C by the length of each timestep, because the one differentiation and the integration exactly cancel out the Jacobians.

        q = self.q
        n = q+1

        C = np.zeros( (n,n), dtype=RTYPE )
        C[0,0] = -1./2.
        C[0,1] =  1./2.
        C[1,0] = -1./2.
        C[1,1] =  1./2.

        if q >= 2:
            t = 1./np.sqrt(6.)
            C[0,2] = -t
            C[1,2] =  t
            C[2,0] =  t
            C[2,1] = -t

        # General formula for C_ji for j,i >= 2.
        for j in range(2,n):
            i = j + 1  # i-1 = j  <=>  i = j+1
            if i >= 2 and i < n:
                C[j,i] =  2. * np.sqrt( 1. / ( ( 2.*j - 1. ) * ( 2.*j + 1. ) ) )
            i = j - 1  # i+1 = j  <=>  i = j-1
            if i >= 2 and i < n:
                C[j,i] = -2. * np.sqrt( 1. / ( ( 2.*j - 1. ) * ( 2.*j - 3. ) ) )

        if self.method == "dG":
            # Jump term for discontinous Galerkin (dG). N[0] is the only basis function that is nonzero at x=-1.
            #
            # See the user manual for derivation.
            #
            C[0,0] += 1.0
        else: # self.method == "cG":
            # In continuous Galerkin (cG), we don't have the jump term; there we must instead:
            #
            #   - Eliminate the first degree of freedom by substituting in its known value (eliminating a row and column, transferring the known terms to the load), or
            #   - Replace the first row with the multiplicative identity and replace the first component of the load with the known value of the first degree of freedom.
            #
            # The first is more elegant (disturbing the matrix structure less), while the second is easier to code. Both approaches achieve the same result.
            # Note that both require implementation also in the timestepping code that fills in the load vector (not just in this helper class).
            #
            # We use the easier second approach, because then we won't need to remap DOF numbers. Also, the  problem is antisymmetric (skew-symmetric),
            # i.e. a symmetric solver could not be used in any case. Note also that the equation system will be rather small, and the hierarchical
            # basis functions are good for numerical linear algebra (far from linear dependence in the sense of the L2 inner product), so there
            # should not be much roundoff or cancellation.
            #
            C[0,:] = 0.0
            C[0,0] = 1.0

        self.C = C

    def __prep_solver(self):
        """Prepare the dG(q) linear equation system for solving."""
        self.LU,self.p = dgesv.lup_packed(self.C)
        self.mincols,self.maxcols = dgesv.find_bands(self.LU, tol=1e-15)  # this allows us to use banded solver (the bw of the LU decomposition of the dG(q) mass matrix is rather small)
#        r   = np.arange(self.q+1, dtype=int)
#        bwL = np.max(r - self.mincols)  # convention: diagonal matrix = zero bandwidth ( https://en.wikipedia.org/wiki/Band_matrix )
#        bwU = np.max(self.maxcols - r)
#        print( "        Banded solver detected bandwidths L=%d (lower bw), U=%d (upper bw)" % (bwL, bwU) )

    def allocate_storage(self, solver_id, n_space_dofs):
        """def allocate_storage(self, solver_id, n_space_dofs):

Allocate and return per-problem storage.

This must be called separately for each problem that will be running simultaneously (even if same settings),
as each problem instance needs its own work space.

This is handled automatically by GalerkinIntegrator.__init__.

Parameters:
    solver_id : hashable
        unique identifier (id(self) at the calling end is a good choice, or if not object-oriented, id(result_array_ww))
    n_space_dofs : int
        number of space DOFs (i.e. size of the 1st-order ODE system) for the problem to be solved.
"""

        assert self.__storage is not None
        assert self.n_time_dofs is not None
        assert self.rule is not None
        assert self.nt_vis is not None

        n_time_dofs = self.n_time_dofs
        n_quad      = self.rule  # Gauss-Legendre has as many quadrature points as the order of the rule being used.
        nt_vis      = self.nt_vis

        cdef DTYPE_t[:,:,::1] g     = np.empty( [n_space_dofs,n_time_dofs,n_quad], dtype=DTYPE, order="C" )  # effective load vector, for each space DOF, for each time DOF, at each integration point
        cdef DTYPE_t[:,::1]   b     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # right-hand sides (integral, over the timestep, of g*psi)
        cdef DTYPE_t[:,::1]   u     = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients (unknowns)
        cdef DTYPE_t[:,::1]   uprev = np.empty( [n_space_dofs,n_time_dofs],        dtype=DTYPE, order="C" )  # Galerkin coefficients from previous iteration
        cdef DTYPE_t[:,::1]   uass  = np.empty( [n_quad,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for integration (this ordering needed for speed!)
        cdef DTYPE_t[::1]     ucorr = np.empty( [n_quad],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for integration)
        cdef DTYPE_t[:,::1]   uvis  = np.empty( [nt_vis,n_space_dofs],             dtype=DTYPE, order="C" )  # u, assembled for visualization
        cdef DTYPE_t[::1]     ucvis = np.empty( [nt_vis],                          dtype=DTYPE, order="C" )  # correction for compensated summation in assemble() (for visualization)
        cdef DTYPE_t[::1]     wrk   = np.empty( [n_space_dofs],                    dtype=DTYPE, order="C" )  # work space for dG(), cG()

        items = {}

        items["n_space_dofs"] = n_space_dofs

        items["g"]     = g
        items["b"]     = b
        items["u"]     = u
        items["uprev"] = uprev
        items["uass"]  = uass
        items["ucorr"] = ucorr
        items["uvis"]  = uvis
        items["ucvis"] = ucvis
        items["wrk"]   = wrk

        self.__storage[solver_id] = items
        return items

    def get_storage(self, solver_id):
        """Return per-problem storage for given solver_id."""
        return self.__storage[solver_id]

    def free_storage(self, solver_id):
        """Deallocate per-problem storage."""
        self.__storage.pop(solver_id, None)   # http://stackoverflow.com/questions/11277432/how-to-remove-a-key-from-a-dictionary


datamanager = None
def init(q=3, method="dG", nt_vis=1, rule=None):
    """def init(q=3, method="dG", nt_vis=1, rule=None):

Initialize Galerkin solver.

This function loads common precalculated data.

This must be called before using the integrators "dG" or "cG".

During one session (run of the process), one initialization is enough for all simulations
using the same method (combination of "q" and "method"). To change the settings, simply initialize again.

The Galerkin solver uses the hierarchical (Lobatto) basis functions (linears, plus bubbles
based on integrated Legendre polynomials).

Parameters (see DataManager.__init__ for full description):
    q : int, >= 1
        Polynomial degree of basis.

    method : str
        Which Galerkin method to use.

    nt_vis : int, >= 1
        Number of visualization points per computed timestep.

         !!! Currently, the value of interp in all subsequent calls to odesolve.ivp() must be the same
             as the value of nt_vis passed here. This is a implementation-technical limitation
             that may be removed in the future. !!!

    rule : int or None
        Order of the Gauss-Legendre quadrature to use for integrating the load vector.
        None = highest available.
"""
    # TODO: relax the technical limitation on nt_vis <-> interp
    global datamanager
    datamanager = DataManager(q, method, nt_vis, rule)




# -*- coding: utf-8 -*-
#
"""Integrate first-order ODE system  u'(t) = f(u, t).

The main point of interest in this library is dG(q), i.e. the
time-discontinuous Galerkin method using a Lobatto basis
(a.k.a. hierarchical polynomial basis). See ivp().

For preparing the data file used by the integrator (pydgq_data.bin),
run the module pydgq.utils.precalc as the main program.

Note also that since the precalc module is not needed once the data file
has been generated, it is not automatically imported.

When this module is imported, it imports all symbols from pydgq.solver.odesolve
into the local namespace.
"""

from __future__ import absolute_import  # https://www.python.org/dev/peps/pep-0328/

__version__ = '0.1.2'

from .solver.odesolve import *




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
from __future__ import division, print_function, absolute_import

from pydgq.solver.kernel_interface cimport CythonKernel
from pydgq.solver.types cimport RTYPE_t

cdef class MyKernel(CythonKernel):
    # custom cdef data attributes go here
    cdef RTYPE_t omega

    # tell Cython that we would like to override callback() with our implementation
    cdef void callback(self, RTYPE_t t) nogil




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Classical implicit integrators to update w by one timestep."""

from __future__ import division, print_function, absolute_import

import numpy as np

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.types import DTYPE, RTYPE
from pydgq.solver.kernel_interface cimport KernelBase
from pydgq.solver.integrator_interface cimport ImplicitIntegrator


cdef class IMR(ImplicitIntegrator):
    def __init__(self, KernelBase rhs, int maxit):
        """def __init__(self, KernelBase rhs, int maxit):

Implicit midpoint rule.

Parameters:
    as in ancestor (pydgq.solver.integrator_interface.ImplicitIntegrator).
"""
        # super
        ImplicitIntegrator.__init__(self, name="IMR", rhs=rhs, maxit=maxit)

        cdef int n_space_dofs = rhs.n
        self.wrk_arr = np.empty( (4*n_space_dofs,), dtype=DTYPE, order="C" )
        self.wrk     = &(self.wrk_arr[0])

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef unsigned int j, m=-1, m2, nequals
        cdef int n_space_dofs = self.rhs.n
        cdef int success = 0

        cdef DTYPE_t* whalf = self.wrk                   # iterative approximation of w(k+1/2)
        cdef DTYPE_t* wp    = &(self.wrk[n_space_dofs])  # iterative approximation of w'(k+1/2)

        # Iterative approximations of w(k+1)
        cdef DTYPE_t* wcur = &(self.wrk[2*n_space_dofs])
        cdef DTYPE_t* wnew = &(self.wrk[3*n_space_dofs])

        cdef RTYPE_t thalf = t + 0.5*dt

        # Trivial initial guess: w(k+1) = w(k)
        #
        # This leads to the *second* iterate being the Forward Euler prediction.
        #
        for j in range(n_space_dofs):
            wcur[j] = w[j]

        # Implicit iteration (Banach fixed point iteration)
        #
        for m in range(self.maxit):
            self.rhs.begin_iteration(m)  # inform RHS kernel that a new iteration starts

            for j in range(n_space_dofs):
                whalf[j] = 0.5 * (w[j] + wcur[j])  # estimate midpoint value of w
            self.rhs.call(whalf, wp, thalf)
            for j in range(n_space_dofs):
                wnew[j] = w[j] + dt*wp[j]  # new estimate of w(k+1)

            # Check convergence; break early if converged to within machine precision.
            #
            # If the timestep is reasonable i.e. small enough to make the truncation error small
            # and actually produce correct results, typically the solution converges to within
            # machine precision in a large majority of cases.
            #
            # If your problem is too large for this to be feasible, set self.maxit accordingly.
            #
            # See
            #
            #  https://randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/
            #  http://stackoverflow.com/questions/3281237/is-the-use-of-machine-epsilon-appropriate-for-floating-point-equality-tests
            #
            nequals = 0
            for j in range(n_space_dofs):
                if wnew[j] == wcur[j]:
                    nequals += 1
                wcur[j] = wnew[j]  # update w(k+1) (old value no longer needed)
            if nequals == n_space_dofs:
                success = 1
                break

        # if timestep failed, try a few more steps with relaxation
        #
        if not success:
            for m2 in range(5):  # we want to be able to return the original value of m
                self.rhs.begin_iteration(self.maxit + m2)  # inform RHS kernel that a new iteration starts

                for j in range(n_space_dofs):
                    whalf[j] = 0.5 * (w[j] + wnew[j])  # estimate midpoint value of w
                self.rhs.call(whalf, wp, thalf)
                for j in range(n_space_dofs):
                    wnew[j] = w[j] + dt*wp[j]  # new estimate of w(k+1)

                nequals = 0
                for j in range(n_space_dofs):
                    if wnew[j] == wcur[j]:
                        nequals += 1
                    wnew[j] = 0.5*wnew[j] + 0.5*wcur[j]  # update w(k+1), with relaxation (old value no longer needed)
                if nequals == n_space_dofs:
                    break

        # w(k+1) computed, update it into w
        for j in range(n_space_dofs):
            w[j] = wnew[j]

        return (m+1)  # Return the number of iterations taken (note that the numbering starts from m=0). If self.maxit iterations were taken, the convergence failed.


cdef class BE(ImplicitIntegrator):
    def __init__(self, KernelBase rhs, int maxit):
        """def __init__(self, KernelBase rhs, int maxit):

Backward Euler (implicit Euler).

Parameters:
    as in ancestor (pydgq.solver.integrator_interface.ImplicitIntegrator).
"""
        # super
        ImplicitIntegrator.__init__(self, name="BE", rhs=rhs, maxit=maxit)

        cdef int n_space_dofs = rhs.n
        self.wrk_arr = np.empty( (3*n_space_dofs,), dtype=DTYPE, order="C" )
        self.wrk     = &(self.wrk_arr[0])

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef unsigned int j, m=-1, m2, nequals
        cdef int n_space_dofs = self.rhs.n
        cdef int success = 0

        cdef DTYPE_t* wp   = self.wrk     # iterative approximation of w'(k+1)

        # Iterative approximations of w(k+1)
        cdef DTYPE_t* wcur = &(self.wrk[n_space_dofs])
        cdef DTYPE_t* wnew = &(self.wrk[2*n_space_dofs])

        cdef RTYPE_t tend = t + dt

        # Trivial initial guess: w(k+1) = w(k)
        #
        # This leads to the *second* iterate being the Forward Euler prediction.
        #
        for j in range(n_space_dofs):
            wcur[j] = w[j]

        # Implicit iteration (Banach fixed point iteration)
        #
        for m in range(self.maxit):
            self.rhs.begin_iteration(m)  # inform RHS kernel that a new iteration starts

            self.rhs.call(wcur, wp, tend)
            for j in range(n_space_dofs):
                wnew[j] = w[j] + dt*wp[j]  # new estimate of w(k+1)

            # Check convergence; break early if converged to within machine precision.
            #
            nequals = 0
            for j in range(n_space_dofs):
                if wnew[j] == wcur[j]:
                    nequals += 1
                wcur[j] = wnew[j]  # update w(k+1) (old value no longer needed)
            if nequals == n_space_dofs:
                success = 1
                break

        # if timestep failed, try a few more steps with relaxation
        #
        if not success:
            for m2 in range(5):  # we want to be able to return the original value of m
                self.rhs.begin_iteration(self.maxit + m2)  # inform RHS kernel that a new iteration starts

                self.rhs.call(wnew, wp, tend)
                for j in range(n_space_dofs):
                    wnew[j] = w[j] + dt*wp[j]  # new estimate of w(k+1)

                nequals = 0
                for j in range(n_space_dofs):
                    if wnew[j] == wcur[j]:
                        nequals += 1
                    wnew[j] = 0.5*wnew[j] + 0.5*wcur[j]  # update w(k+1), with relaxation (old value no longer needed)
                if nequals == n_space_dofs:
                    break

        # w(k+1) computed, update it into w
        for j in range(n_space_dofs):
            w[j] = wnew[j]

        return (m+1)  # Return the number of iterations taken (note that the numbering starts from m=0). If self.maxit iterations were taken, the convergence failed.




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Interface for evaluating the RHS f() in  w' = f(w, t).

These base classes connect the IVP solver with the user-provided custom code
for computing the RHS. Cython and Python interfaces are available.

For a specific, arbitrary nonlinear problem, a custom kernel must be provided.
This can be done in either of two ways:

    - a) implement a cdef class
        - inherit from pydgq.solver.kernel_interface.CythonKernel
        - override __init__(), add any needed parameters, call CythonKernel.__init__(self, n)
        - override cdef callback(...) to compute w' for your RHS

    - b) implement a regular Python class
        - inherit from pydgq.solver.kernel_interface.PythonKernel
        - override __init__(), add any needed parameters, call PythonKernel.__init__(self, n)
        - override def callback(...) to compute w' for your RHS

Cython kernels run in nogil mode, allowing parallel processing of several independent problems.
Useful in use cases with many independent small problems, and for OpenMP based parallel processing
inside callback() (using cython.parallel.prange()).

Python kernels acquire the GIL for each call of callback(). Useful for quick prototyping
of new ODE systems, and for large problems where the Python function call overhead is not significant.

See pydgq.solver.builtin_kernels for some Cython-based example kernels.
"""

from __future__ import division, print_function, absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t


###############
# Base classes
###############

cdef class KernelBase:
#    cdef DTYPE_t* w    # old state vector (memory owned by caller)
#    cdef DTYPE_t* out  # new state vector (memory owned by caller)
#    cdef int n         # n_space_dofs
#    cdef int timestep
#    cdef int iteration

    def __init__(self, int n):
        """def __init__(self, int n):

Base class for all kernels.

Do not inherit directly from this; instead, see CythonKernel and PythonKernel
depending on which language you wish to implement your kernel in.

Parameters:
    n : int
        Number of DOFs of the ODE system.

Data attributes:
    w : DTYPE_t*
        input, old state vector (memory owned by caller)
    out : DTYPE_t*
        output, new state vector (memory owned by caller)
    n : int
        number of DOFs (taken from __init__)
    timestep : int
        number of current timestep
    iteration : int
        number of current nonlinear (Banach/Picard) iteration
"""
        self.n = n

    # The solver (odesolve.ivp()) calls begin_timestep() when it begins a new timestep.
    #
    # Each integrator implementation calls begin_iteration() when it begins a new Banach/Picard iteration.
    # (Explicit integrators also call it once, for "iteration 0".)
    #
    # This metadata is intended for the actual computational kernel; the kernel base classes do not need it.
    #
    # timestep  : 0-based, 0 = initial condition, 1 = first timestep, 2 = second timestep, ...
    # iteration : 0-based. Special value -1 = evaluation of final result from this timestep. (Used if saving also w' in the results arrays.)
    #
    cdef void begin_timestep(self, int timestep) nogil:
        self.timestep  = timestep
    cdef void begin_iteration(self, int iteration) nogil:
        self.iteration = iteration

    # The call interface. The solver calls this when it wants to evaluate w'.
    #
    # Implemented in derived classes.
    #
    cdef void call(self, DTYPE_t* w, DTYPE_t* out, RTYPE_t t) nogil:
        pass


cdef class CythonKernel(KernelBase):
    # we override __init__ only in order to provide a docstring.
    def __init__(self, int n):
        """def __init__(self, int n):

Base class for kernels implemented in Cython.

Cython kernels will run in nogil mode.

No further docstrings are provided, because the rest of this class
is not visible from the Python level. See the source code in
pydgq.solvers.kernel_interface.pyx for details.

Basically, in your cdef class, override the method

    cdef void callback(self, RTYPE_t t) nogil:

The necessary arrays can be accessed as self.w and self.out.
Both arrays have self.n elements.

If callback() is not overridden, this class implements a no-op kernel: w' = 0.
"""
        KernelBase.__init__(self, n)

    # Implementation of call() for Cython kernels.
    #
    cdef void call(self, DTYPE_t* w, DTYPE_t* out, RTYPE_t t) nogil:
        self.w   = w
        self.out = out
        self.callback(t)

    # Hook for custom code.
    #
    # Default no-op kernel: w' = 0
    #
    # Override this method in derived classes to provide your computational kernel.
    #
    cdef void callback(self, RTYPE_t t) nogil:
        cdef int j
        for j in range(self.n):
            self.out[j] = 0.0


cdef class PythonKernel(KernelBase):
#    cdef DTYPE_t[::1] w_arr
#    cdef DTYPE_t[::1] out_arr

    # we override __init__ only in order to provide a docstring.
    def __init__(self, int n):
        """def __init__(self, int n):

Base class for kernels implemented in pure Python.

Python kernels will acquire the GIL for calling callback(),
which is a regular Python-level method (in Cython parlance,
"def method").

See callback().

**Cython-level** data attributes added by PythonKernel:
    w_arr : DTYPE_t[::1] (visible from Python as "w")
        Python-accessible view of Cython-level self.w
    out_arr : DTYPE_t[::1] (visible from Python as "out")
        Python-accessible view of Cython-level self.out
"""
        KernelBase.__init__(self, n)

    # Implementation of call() for Python kernels.
    #
    cdef void call(self, DTYPE_t* w, DTYPE_t* out, RTYPE_t t) nogil:
        self.w   = w
        self.out = out
        with gil:
            self.w_arr   = <DTYPE_t[:self.n:1]>w
            self.out_arr = <DTYPE_t[:self.n:1]>out
            self.callback(t)

    # This wrapper is needed because cdef data attributes of cdef classes
    # live in a C struct (which is invisible from the Python level).
    #
    def __getattr__(self, name):
        """def __getattr__(self, name):

Provide Python-level access to Cython data attributes.

The names are the same as the Cython counterparts,
except w_arr and out_arr, which are named w and out.
"""
        if name == "n":
            return self.n
        elif name == "w":
            return self.w_arr
        elif name == "out":
            return self.out_arr
        elif name == "timestep":
            return self.timestep
        elif name == "iteration":
            return self.iteration
        else:
            raise AttributeError("No such attribute '%s'" % (name))

    def callback(self, RTYPE_t t):
        """def callback(self, RTYPE_t t):

Python-based callback (hook for custom code).

Override this method in derived classes to provide your computational kernel.

Use self.w and self.out normally; in PythonKernel, when used from the Python level,
these are Python-accessible views to the underlying memory.

Both arrays have self.n elements.

If callback() is not overridden, this class implements a no-op kernel: w' = 0.
"""
        cdef int j
        for j in range(self.n):
            self.out[j] = 0.0




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#

from __future__ import division, print_function, absolute_import

from pydgq.solver.types cimport RTYPE_t, ZTYPE_t

# Compensated summation (Kahan summation), as accumulation to a sum variable.
#
# For operands of wildly varying magnitudes, this helps accuracy a lot,
# at the cost of four times the arithmetic.
#
# Note that this only has any effect on accuracy when there are at least three terms to be summed.
# For two terms, the correction c will be computed, but will not be used to correct the sum.
#
# See:
#     http://en.wikipedia.org/wiki/Kahan_summation_algorithm
#
# Parameters:
#     s :       in/out, accumulated sum, RTYPE_t* pointing to a single RTYPE_t
#     c :       in/out, accumulated correction, RTYPE_t* pointing to a single RTYPE_t
#     operand : in,     the new term to add to the sum, RTYPE_t
#
# We use pointers only for the ability to use "in/out" parameters.
#
# Usage:
#
#     # To initialize, set s and c to zero.
#     cdef RTYPE_t s = 0.0
#     cdef RTYPE_t c = 0.0
#
#     # Loop over the terms of the sum.
#     cdef int k
#     for k in range(10):
#       accumulate( &s, &c, my_data[k] )  # here my_data is an array of RTYPE_t containing the numbers to be summed.
#     # now:
#     #  - s is the result of the sum
#     #  - c contains the final correction term (that is too small to be representable in s)
#
# Or, slightly optimized (may matter with small arrays):
#
#     cdef RTYPE_t s = my_data[0]  # Each term in the sum is RTYPE_t, so any one of them is exactly representable.
#     cdef RTYPE_t c = 0.0         # Correction is zero at the beginning.
#
#     # Loop over the terms of the sum, starting from the second one.
#     cdef int k
#     for k in range(1,10):
#       accumulate( &s, &c, my_data[k] )
#     # now:
#     #  - s is the result of the sum
#     #  - c contains the final correction term (that is too small to be representable in s)
#
cdef inline void accumulate( RTYPE_t* s, RTYPE_t* c, RTYPE_t operand ) nogil:
    # In compensated summation (Kahan summation),
    #
    #   s += x
    #
    # becomes:
    #
    #   y = x - c          # Subtract the old correction. This works because typically |x| << |s|.
    #   t = s + y          # Accumulate y, but don't overwrite yet, we still need the old value of s to compute the new correction.
    #   c = (t - s) - y    # Update the correction: (new - old) values of the sum vs. what was actually added.
    #   s = t              # Update s.
    #
    # where s and c are initially zero (before the accumulation of the first term of the sum).
    #
    cdef RTYPE_t y = operand - c[0]
    cdef RTYPE_t t = s[0] + y
    c[0] = (t - s[0]) - y
    s[0] = t


# Version for complex numbers.
#
cdef inline void accumulatez( ZTYPE_t* s, ZTYPE_t* c, ZTYPE_t operand ) nogil:
    cdef ZTYPE_t y = operand - c[0]
    cdef ZTYPE_t t = s[0] + y
    c[0] = (t - s[0]) - y
    s[0] = t


# Cumulative summation for 1D data (like np.cumsum, but with compensated summation).
#
# See compsum.pyx for an explanation of the parameters.
#
cdef void cs1dr( RTYPE_t* data, RTYPE_t* out, unsigned int n ) nogil
cdef void cs1dz( ZTYPE_t* data, ZTYPE_t* out, unsigned int n ) nogil

